"""
Module to process tracking log files from edX
"""
import errno
import glob
import gzip
import json
import os
import signal
import sys
import time
import traceback
from datetime import datetime
from json.decoder import JSONDecodeError
from multiprocessing.pool import (
    Pool, TimeoutError
)
from typing import Dict, List, Union

from dateutil.parser import parse as parse_date

from simeon.download import utilities as utils
from simeon.exceptions import (
    EarlyExitError, MissingSchemaException, SplitException,
)
from simeon.report import utilities as rutils


SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)


def _process_initializer():
    """
    Process initializer for multiprocessing pools
    """
    def sighandler(sig, frame):
        raise EarlyExitError(
            'Tracking log splitting interrupted prematurely'
        )
    sigs = [signal.SIGABRT, signal.SIGTERM, signal.SIGINT]
    for sig in sigs:
        signal.signal(sig, signal.SIG_DFL)


def _cleanup_handles(fhandles, sleep=1):
    """
    Flush and close the given file handles.
    Sleep sleep number of seconds if necessary,
    so the OS can reclaim the file descriptors
    associated with the given handles.
    """
    for fhandle in fhandles:
        fhandle.flush()
        os.fsync(fhandle.fileno())
        os.close(fhandle.fileno())
    if sleep:
        time.sleep(sleep)


# pylint:disable=unsubscriptable-object
def process_line(
    line: Union[str, bytes], lcount: int,
    date: Union[None, datetime]=None, is_gzip=True,
    courses: List[str]=None
) -> dict:
    """
    Process the line from a tracking log file and return the reformatted
    line (deserialized) along with the name of its destination file.

    :type line: Union[str, bytes]
    :param line: A line from the tracking logs
    :type lcount: int
    :param lcount: The line number of the given line
    :type date: Union[None, datetime]
    :param date: The date of the file where this line comes from.
    :type is_gzip: bool
    :param is_gzip: Whether or not this line came from a GZIP file
    :type courses: Union[Iterable[str], None]
    :param courses: A list of course IDs whose records are exported
    :rtype: Dict[str, Union[Dict[str, str], str]]
    :return: Dictionary with both the data and its destination file name
    """
    line = line.strip()
    if isinstance(line, bytes):
        line = line.decode('utf8', 'ignore')
    if not line.startswith('{'):
        if 'localhost {' in line[:27]:
            line = line.split('localhost ')[-1]
    try:
        record = json.loads(line)
    except (JSONDecodeError, TypeError):
        return {
            'data': line,
            'filename': os.path.join(
                'dead_letters',
                'dead_letter_queue_{p}.json.gz'.format(p=os.getpid())
            )
        }
    if not isinstance(record.get('event'), dict):
        try:
            record['event'] = json.loads(record.get('event', '{}'))
        except (JSONDecodeError, TypeError):
            record['event'] = {'event': record['event']}
    course_id = utils.get_course_id(record)
    if courses and course_id not in courses:
        return {}
    record['course_id'] = course_id
    try:
        utils.rephrase_record(record)
    except KeyError:
        return {
            'data': line,
            'filename': os.path.join(
                'dead_letters',
                'dead_letter_queue_{p}.json.gz'.format(p=os.getpid())
            )
        }
    if any(k not in record for k in ('event', 'event_type')):
        return {
            'data': line,
            'filename': os.path.join(
                'dead_letters',
                'dead_letter_queue_{p}.json.gz'.format(p=os.getpid())
            )
        }
    if not date:
        try:
            date = parse_date(record.get('time', ''))
            outfile = utils.make_tracklog_path(
                course_id, date.strftime('%Y-%m-%d'), is_gzip
            )
        except Exception:
            ext = '.gz' if is_gzip else ''
            outfile = os.path.join(
                course_id.replace('.', '_').replace('/', '__') or 'UNKNOWN',
                'tracklog-unknown.json{x}'.format(x=ext)
            )
    else:
        outfile = utils.make_tracklog_path(
            course_id, date.strftime('%Y-%m-%d'), is_gzip
        )
    return {'data': record, 'filename': outfile}


# pylint:enable=unsubscriptable-object
def split_tracking_log(
    filename: str, ddir: str, dynamic_date: bool=False,
    courses: List[str]=None,
    schema_dir=SCHEMA_DIR,
):
    """
    Split the records in the given GZIP tracking log file.
    This function is very resource hungry because it keeps around
    a lot of open file handles and writes to them whenever it processes
    a good record. Some attempts are made to keep records around whenever
    the process is no longer allowed to open new files. But that will likely
    lead to the exhaustion of the running process's alotted memory.

    :NOTE: If you've got a better way, please update me.

    :type filename: str
    :param filename: The GZIP file to split
    :type ddir: str
    :param ddir: Destination directory of the generated file
    :type dynamic_date: bool
    :param dynamic_date: Use dates from the JSON records to make
        output file names
    :type courses: Union[Iterable[str], None]
    :param courses: A list of course IDs whose records are exported
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where to find schema files
    :rtype: bool
    :return: True if files have been generated. False, otherwise
    """
    schema_dir = schema_dir or SCHEMA_DIR
    targets = glob.iglob(os.path.join(
        schema_dir, '*tracking_log.json'
    ))
    schema_file = next(targets, None)
    if schema_file is None or not os.path.exists(schema_file):
        msg = (
            'No valid schema file tracking_log.json was found '
            'in the schema directory {d}. Please provide a valid '
            'schema directory.'
        )
        raise MissingSchemaException(
            'The schema file {f} does not exist. '
            'Please provide a valid schema directory'.format(f=schema_file)
        )
    with open(schema_file) as sfh:
        schema = json.load(sfh).get('tracking_log')
    if not isinstance(courses, set):
        courses = set(c for c in (courses or []))
    fhandles = dict()
    if not dynamic_date:
        date = utils.get_file_date(filename)
    else:
        date = None
    with gzip.open(filename) as zfh:
        stragglers = []
        for i, line in enumerate(zfh):
            line_info = process_line(line, i + 1, date=date, courses=courses)
            if not line_info:
                continue
            data = line_info['data']
            user_id = (data.get('context') or {}).get('user_id')
            username = data.get('username')
            if not any([user_id, username]):
                continue
            if isinstance(data, dict):
                if courses and data.get('course_id') not in courses:
                    continue
            fname = line_info.get('filename')
            fname = os.path.join(ddir, fname)
            if fname not in fhandles:
                try:
                    fhandles[fname] = utils.make_file_handle(
                        fname, is_gzip=True
                    )
                except OSError as excp:
                    if excp.errno == errno.EMFILE:
                       stragglers.append(line_info)
                       continue
                    if excp.errno == errno.ENAMETOOLONG:
                        continue
                    raise excp
            fhandle = fhandles[fname]
            if not isinstance(data, str):
                rutils.check_record_schema(data, schema)
                rutils.drop_extra_keys(data, schema)
                data = json.dumps(data)
            if isinstance(fhandle, gzip.GzipFile):
                fhandle.write(data.encode('utf8', 'ignore') + b'\n')
            else:
                fhandle.write(data + '\n')
        if not stragglers:
            return bool(fhandles)
        # Working around EMFILE errors
        _cleanup_handles(fhandles.values())
        # Sort the stragglers by file name and use a file tracker
        stragglers = sorted(stragglers, key=lambda s: s.get('filename'))
        pfname = None
        while stragglers:
            try:
                rec = stragglers.pop()
                fname = rec.get('filename')
                fname = os.path.join(ddir, fname)
                if pfname and pfname != fname:
                    try:
                        _cleanup_handles([fhandles[pfname]], None)
                    except OSError:
                        pass
                data = line_info['data']
                if fname not in fhandles:
                    try:
                        fhandle = utils.make_file_handle(fname, is_gzip=True)
                        fhandles[fname] = fhandle
                    except OSError as excp:
                        if excp.errno == errno.EMFILE:
                            stragglers.append(rec)
                            continue
                        if excp.errno == errno.ENAMETOOLONG:
                            continue
                        raise excp
                if not isinstance(data, str):
                    rutils.check_record_schema(data, schema)
                    rutils.drop_extra_keys(data, schema)
                    data = json.dumps(data)
                if isinstance(fhandle, gzip.GzipFile):
                    fhandle.write(data.encode('utf8', 'ignore') + b'\n')
                else:
                    fhandle.write(data + '\n')
                pfname = fname
            except IndexError:
                break
    return bool(fhandles)


def batch_split_tracking_logs(
    filenames, ddir, dynamic_date=False,
    courses=None, verbose=True, logger=None,
    size=10, schema_dir=SCHEMA_DIR, debug=False,
):
    """
    Call split_tracking_log on each file inside a process or thread pool
    """
    schema_dir = schema_dir or SCHEMA_DIR
    if not size or size > len(filenames):
        size = len(filenames)
    splits = 0
    processed = 0
    with Pool(size, initializer=_process_initializer) as pool:
        results = dict()
        for fname in filenames:
            if verbose and logger:
                logger.info('Splitting {f}'.format(f=fname))
            result = pool.apply_async(
                func=split_tracking_log,
                kwds=dict(
                    filename=fname, ddir=ddir,
                    dynamic_date=dynamic_date, courses=courses,
                )
            )
            results[fname] = (result, False)
        while processed < len(filenames):
            for fname in results:
                result, done = results[fname]
                if done:
                    continue
                try:
                    # Do the rounds every minute and check if a worker
                    # is done with its job
                    rc = result.get(timeout=60)
                    splits += rc
                    results[fname] = (result, True)
                    processed += 1
                    if rc:
                        if verbose and logger:
                            logger.info('Done splitting {f}'.format(f=fname))
                        continue
                    if logger:
                        errmsg = (
                            'No files were extracted while splitting the tracking '
                            'log file {f!r} with the given criteria. Moving on...'
                        )
                        logger.warning(errmsg.format(f=fname))
                        logger.warning('Done splitting {f}'.format(f=fname))
                except TimeoutError:
                    continue
                except KeyboardInterrupt:
                    msg = 'Failed to split {f}: Interrupted by the user'
                    logger.error(msg.format(f=fname))
                    return False
                except:
                    _, excp, tb = sys.exc_info()
                    msg = 'Failed to split {f}{e}'
                    exit_excepts = (EarlyExitError, SystemExit)
                    if isinstance(excp, exit_excepts):
                        raise SplitException(
                            msg.format(f=fname, e='')
                        )
                    else:
                        if debug:
                            traces = [': {e}'.format(e=excp)]
                            traces += map(str.strip, traceback.format_tb(tb))
                            excp_str = '\n'.join(traces)
                        logger.error(msg.format(f=fname, e=excp_str))
                    results[fname] = (result, True)
                    processed += 1
    return splits == len(filenames)



"""
Utility functions and classes to help with downloading and decrypting
the data from S3.
"""
from .aws import (
    S3Blob, make_s3_bucket,
)
from .emails import (
    process_email_file,
)
from .logs import (
    batch_split_tracking_logs, split_tracking_log,
)
from .sqls import (
    process_sql_archive,
)
from .utilities import (
    decrypt_files,
)



"""
Utilities functions and classes to help with loading data to Google Cloud
"""
import glob
import gzip
import os
from datetime import datetime
from typing import List

from google.cloud import (
    bigquery, exceptions as gcp_exceptions, storage
)

from simeon.upload import utilities as uputils
from simeon.report import utilities as rutils
from simeon.exceptions import LoadJobException
from simeon.upload.utilities import (
    SCHEMA_DIR
)


FILE_FORMATS = {
    'log': ['json'],
    'sql': ['csv', 'txt', 'sql', 'json']
}
MERGE_DDL = """MERGE {first} f USING {second} s
ON f.{column} = s.{column}
WHEN NOT MATCHED THEN
INSERT ROW
"""
DST_DESC = {
    'log': 'Dataset to host the tracking log data from edX courses',
    'sql': (
        'Dataset to host all the tables that are computed from a combination'
        ' of tab-delimited files from edX\'s weekly SQL data dump and '
        'tracking log tables.'
    ),
    'email': (
        'Dataset to host the dimensional details about users\' email'
        ' adresses and email opt-in preferences'
    ),
}


class BigqueryClient(bigquery.Client):
    """
    Subclass bigquery.Client and add convenience methods
    """
    def load_tables_from_dir(
        self, dirname: str, file_type: str, project: str,
        create: bool, append: bool, use_storage: bool=False,
        bucket: str=None, max_bad_rows=0,
        schema_dir=SCHEMA_DIR, format_='json', patch=False,
    ) -> List[bigquery.LoadJob]:
        """
        Load all the files in the given directory.

        :type dirname: str
        :param dirname: Grandparent or parent directory of split up files
        :type file_type: str
        :param file_type: One of sql, email, log, rdx
        :type project: str
        :param project: Target GCP project
        :type create: bool
        :param create: Whether or not to create the destination table
        :type append: bool
        :param append: Whether or not to append the records to the table
        :type use_storage: bool
        :param use_storage: Whether or not to load the data from GCS
        :type bucket: str
        :param bucket: GCS bucket name to use
        :type max_bad_rows: int
        :param max_bad_rows: Max number of bad rows allowed during loading
        :type schema_dir: str
        :param schema_dir: Directory where schema files are found
        :type format_: str
        :param format_: File format (json or csv)
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :rtype: List[bigquery.LoadJob]
        :returns: List of load jobs
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        formats = FILE_FORMATS.get(file_type, [])
        patts = (
            os.path.join(dirname, '*.{f}*.gz'),
            os.path.join(dirname, '*', '*.{f}*.gz')
        )
        files = []
        for patt in patts:
            for format_ in formats:
                files.extend(glob.glob(patt.format(f=format_)))
        jobs = []
        for file_ in files:
            jobs.append(
                self.load_one_file_to_table(
                    file_, file_type, project, create, append,
                    use_storage, bucket, max_bad_rows, schema_dir, format_,
                    patch
                )
            )
        return jobs

    def load_one_file_to_table(
        self, fname: str, file_type: str, project: str,
        create: bool, append: bool, use_storage: bool=False,
        bucket: str=None, max_bad_rows=0,
        schema_dir=SCHEMA_DIR, format_='json', patch=False,
    ):
        """
        Load the given file to a target BigQuery table

        :type fname: str
        :param fname: The specific file to load
        :type file_type: str
        :param file_type: One of sql, email, log, rdx
        :type project: str
        :param project: Target GCP project
        :type create: bool
        :param create: Whether or not to create the destination table
        :type append: bool
        :param append: Whether or not to append the records to the table
        :type use_storage: bool
        :param use_storage: Whether or not to load the data from GCS
        :type bucket: str
        :param bucket: GCS bucket name to use
        :type max_bad_rows: int
        :param max_bad_rows: Max number of bad rows allowed during loading
        :type schema_dir: Union[None, str]
        :param schema_dir: Directory where schema files are found
        :type format_: str
        :param format_: File format (json or csv)
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :rtype: bigquery.LoadJob
        :returns: The LoadJob object associated with the work being done
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        use_storage = use_storage or fname.startswith('gs://')
        if use_storage:
            if bucket is None:
                raise ValueError('use_storage=True requires a bucket name')
            loader = self.load_table_from_uri
        else:
            loader = self.load_table_from_file
        job_prefix = '{t}_data_load_{dt}-'.format(
            t=file_type, dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
        )
        dest = uputils.local_to_bq_table(fname, file_type, project)
        # dataset = self.dataset(dest.rsplit('.', 1)[0])
        dataset = bigquery.Dataset.from_string(dest.rsplit('.', 1)[0])
        dataset.description = DST_DESC.get(file_type)
        self.create_dataset(dataset, exists_ok=True)
        if use_storage:
            if not fname.startswith('gs://'):
                fname = uputils.local_to_gcs_path(fname, file_type, bucket)
        else:
            fname = gzip.open(fname, 'rb')
        config, desc = uputils.make_bq_load_config(
            table=dest, schema_dir=schema_dir,
            append=append, create=create, file_format=format_,
            max_bad_rows=max_bad_rows
        )
        dest = bigquery.Table.from_string(dest)
        dest.description = desc
        if patch:
            # Update the destination table's description.
            # Catch the error raised when the destination table
            # does not exist.
            try:
                self.update_table(dest, ['description'])
            except gcp_exceptions.NotFound:
                pass
        return loader(
            fname, dest, job_config=config, job_id_prefix=job_prefix
        )

    @staticmethod
    def extract_error_messages(errors):
        """
        Return the error messages from given list of error objects (dict)
        """
        messages = []
        if isinstance(errors, dict):
            members = []
            for e in errors.values():
                members += e
        else:
            members = errors
        for err in members:
            msg = err.get('message', '')
            if not msg:
                continue
            src = err.get('source', '')
            if src:
                msg = 'Source: {s} - {m}'.format(s=src, m=msg)
            loc = err.get('location', '')
            if loc:
                msg = '{m} - File: {f}'.format(m=msg, f=loc)
            messages.append(msg)
        return messages

    def merge_to_table(
        self, fname, table, col,
        schema_dir=SCHEMA_DIR, use_storage=False, patch=False,
    ):
        """
        Merge the given file to the target table name.
        If the latter does not exist, create it first.
        This process waits for all the jobs needed

        :type fname: str
        :param fname: A local file name or a GCS URI
        :type table: str
        :param table: Fully qualified BigQuery table name
        :type col: str
        :param col: Column by which to merge
        :type schema_dir: Union[None, str]
        :param schema_dir: The directory where schema files live
        :type use_storage: bool
        :param use_storage: Whether or not the given path is a GCS URI
        :type patch: bool
        :param patch: Whether or not to patch the description of the table
        :rtype: bigquery.QueryJob
        :returns: The QueryJob object associated with the merge carried out
        :raises: Propagates everything from the underlying package
        """
        schema_dir = schema_dir or SCHEMA_DIR
        if len(table.split('.')) < 3:
            table = '{p}.{t}'.format(p=self.project, t=table)
        dataset = table.rsplit('.', 1)[0]
        bqtable = bigquery.Table.from_string(table)
        temp_table = bigquery.Table.from_string(table + '_temp')
        schema, desc = uputils.get_bq_schema(table, schema_dir=schema_dir)
        bqtable.schema = schema
        bqtable.description = desc
        temp_table.schema = schema
        if patch:
            # Update the destination table's description.
            # Catch the error raised when the destination table
            # does not exist.
            try:
                self.update_table(bqtable, ['description'])
            except gcp_exceptions.NotFound:
                pass
        temp_table.description = desc
        job_prefix = '{t}_data_load_{dt}-'.format(
            t=bqtable.table_id,
            dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
        )
        config, _ = uputils.make_bq_load_config(
            table=table, schema_dir=schema_dir, append=False,
            create=True, file_format='json',
        )
        self.create_dataset(dataset, exists_ok=True)
        for tbl in (bqtable, temp_table):
            self.create_table(tbl, exists_ok=True)
        if use_storage:
            loader = self.load_table_from_uri
        else:
            loader = self.load_table_from_file
            fname = gzip.open(fname, 'rb')
        job = loader(
            fname, temp_table, job_config=config, job_id_prefix=job_prefix
        )
        rutils.wait_for_bq_job_ids([job.job_id], self)
        if job.errors:
            msg = 'Merge job failed with: {e}'
            raise LoadJobException(msg.format(
                e='\n'.join(self.extract_error_messages(job.errors))
            ))
        query = MERGE_DDL.format(
            first=table, second=table + '_temp', column=col,
        )
        qjob = self.query(query)
        rutils.wait_for_bq_job_ids([qjob.job_id], self)
        if qjob.errors:
            msg = 'Merge job failed with: {e}'
            raise LoadJobException(msg.format(
                e='\n'.join(self.extract_error_messages(job.errors))
            ))
        self.delete_table(temp_table, not_found_ok=True)


class GCSClient(storage.Client):
    """
    Make a client to load data files to GCS
    """
    def load_one_file_to_gcs(
        self, fname: str, file_type: str, bucket: str
    ):
        """
        Load the given file to GCS

        :type fname: str
        :param fname: The local file to load to GCS
        :type file_type: str
        :param: file_type: One of sql, email, log, rdx
        :type bucket: str
        :param bucket: GCS bucket name
        :type overwrite: bool
        :param overwrite: Overwrite the target blob if it exists
        :rtype: None
        :returns: Nothing, but should load the given file to GCS
        :raises: Propagates everything from the underlying package
        """
        dest = storage.Blob.from_string(
            uputils.local_to_gcs_path(fname, file_type, bucket),
            client=self
        )
        if 'cold' in file_type.lower():
            dest.storage_class = storage.constants.COLDLINE_STORAGE_CLASS
        dest.upload_from_filename(fname, timeout=20 * 60)

    def load_dir(
        self, dirname: str, file_type: str, bucket: str
    ):
        """
        Load all the files in the given directory or
        any immediate subdirectories

        :type dirname: str
        :param dirname: The directory whose files are loaded
        :type file_type: str
        :param: file_type: One of sql, email, log, rdx
        :type bucket: str
        :param bucket: GCS bucket name
        :rtype: None
        :returns: Nothing, but should load file(s) in dirname to GCS
        :raises: Propagates everything from the underlying package
        """
        formats = FILE_FORMATS.get(file_type, [])
        patts = (
            os.path.join(dirname, '*.{f}*.gz'),
            os.path.join(dirname, '*', '*.{f}*.gz')
        )
        files = []
        for patt in patts:
            for format_ in formats:
                files.extend(glob.glob(patt.format(f=format_)))
        for fname in files:
            self.load_one_file_to_gcs(fname, file_type, bucket)



"""
Integration tests for the simeon CLI tool.
These can be quite expensive since they read from and write to files.
"""
import os
import subprocess as sb
import unittest


def _run_command(cmd):
    """
    Run the given command and return the generated Popen object
    """
    proc = sb.Popen(cmd.split(), stderr=sb.PIPE, stdout=sb.PIPE)
    proc.wait()
    return proc


class SimeonCLIIntegration(unittest.TestCase):
    """
    Test the simeon CLI command
    """
    def setUp(self):
        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.log_file = os.path.join(
            self.this_dir, 'fixtures', 'file.log'
        )
        self.config_file = os.path.join(
            self.this_dir, 'fixture', 'config.ini'
        )
        self.good_subcommands = [
            'download', 'list', 'push', 'split'
        ]
        self.bad_subcommands = [
            'upload', 'inspect', 'run', 'stupid', 'spli', 'donwland',
            'psh', 'splt', 'lst', 'upoad', 'uload'
        ]
        self.global_options = [
            '--log-file {f}'.format(f=self.log_file),
            '--config-file {f}'.format(f=self.config_file),
            '--quiet', '--debug'
        ]
        self.proc = None

    def tearDown(self):
        if self.proc:
            if self.proc.stdout:
                self.proc.stdout.close()
            if self.proc.stderr:
                self.proc.stderr.close()
        for file_ in (self.log_file, self.config_file):
            try:
                os.remove(file_)
            except OSError:
                continue

    def test_good_subcommands(self):
        """
        Test that the subcommands have not been changed.
        """
        for subcommand in self.good_subcommands:
            cmd = 'simeon {sub} --help'.format(sub=subcommand)
            with self.subTest('Running {cmd}'.format(cmd=cmd)):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_bad_subcommands(self):
        """
        Test that mispelled subcommands should fail
        """
        for subcommand in self.bad_subcommands:
            cmd = 'simeon {sub} --help'.format(sub=subcommand)
            with self.subTest('Running {cmd}'.format(cmd=cmd)):
                self.proc = _run_command(cmd)
                self.assertNotEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_global_options(self):
        """
        Test that the global options have not changed
        """
        for option in self.global_options:
            cmd = 'simeon {o} --help'.format(o=option)
            msg = 'Checking that global option {o} exists'.format(o=option)
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()


class SimeonGeoIPCLIIntegration(unittest.TestCase):
    """
    Test the simeon-geoip CLI tool
    """
    def setUp(self):
        self.this_dir = os.path.dirname(os.path.abspath(__file__))
        self.log_file = os.path.join(
            self.this_dir, 'fixtures', 'file.log'
        )
        self.config_file = os.path.join(
            self.this_dir, 'fixtures', 'config.ini'
        )
        self.un_file = os.path.join(
            self.this_dir, 'fixtures', 'un_data.csv'
        )
        self.output_file = os.path.join(
            self.this_dir, 'fixtures', 'geoip.json.gz'
        )
        self.global_options = [
            '--log-file {f}'.format(f=self.log_file),
            '--config-file {f}'.format(f=self.config_file),
            '--quiet',
        ]
        self.extract_good_options = [
            '--un-data {f}'.format(f=self.un_file),
            '--output {f}'.format(f=self.output_file),
            '--quiet', '--tracking-logs',
        ]
        self.merge_good_options = [
            '--geo-table dataset.table',
            '--project gcp-project-id',
            '--service-account-file saccount_file.json',
            '--column ip',
        ]
        self.proc = None

    def tearDown(self):
        if self.proc:
            streams = (self.proc.stdout, self.proc.stderr)
            for stream in streams:
                if stream:
                    stream.close()
        for file_ in (self.log_file, self.un_file, self.output_file):
            try:
                os.remove(file_)
            except OSError:
                continue

    def test_geoip_global_options(self):
        """
        Test that global options for simeon-geoip have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip {o} --help'.format(o=option)
            msg = 'Checking that option {o} for simeon-geoip exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_geoip_extract_options(self):
        """
        Test that options for simeon-geoip extract have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip extract {o} --help'.format(o=option)
            msg = 'Checking that option {o} for extract exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()

    def test_geoip_merge_options(self):
        """
        Test that options for simeon-geoip merge have not changed.
        """
        for option in self.global_options:
            cmd = 'simeon-geoip merge {o} --help'.format(o=option)
            msg = 'Checking that option {o} for merge exists'.format(
                o=option
            )
            with self.subTest(msg):
                self.proc = _run_command(cmd)
                self.assertEqual(self.proc.returncode, 0)
                self.tearDown()


if __name__ == '__main__':
    unittest.main()



"""
Utility functions and classes associated with uploading data to GCP, so far.
"""
import glob
import json
import os
import re

from google.cloud import bigquery
from google.cloud import storage

from simeon.exceptions import (
    BigQueryNameException, MissingSchemaException
)


SEGMENTS = {
    'log': 'TRACKING-LOGS',
    'email': 'EMAIL',
    'sql': 'SQL',
    'rdx': 'RDX',
    'cold': 'COLD',
}
SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'schemas'
)
SCHEMAS = {}


def course_to_gcs_folder(course_id: str, file_type: str, bucket: str) -> str:
    """
    Use the given course ID to make a Google Cloud Storage path

    :type course_id: str
    :param course_id: edX course ID to format into a GCS path
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type bucket: str
    :param bucket: A GCS bucket name
    :rtype: str
    :return: A nicely formatted GCS path
    """
    segment = SEGMENTS.get(file_type)
    if not segment:
        raise ValueError(
            'file_type is not one of these: {ft}'.format(
                ft=', '.join(SEGMENTS)
            )
        )
    if not bucket.startswith('gs://'):
        bucket = 'gs://{b}'.format(b=bucket)
    dirname = course_id.replace('/', '__').replace('.', '_')
    return '{b}/{s}/{d}'.format(b=bucket, s=segment, d=dirname)


def local_to_gcs_path(fname: str, file_type: str, bucket: str) -> str:
    """
    Convert the local file name into a GCS path

    :type fname: str
    :param fname: A local file name
    :type file_type: str
    :param file_type: One of sql, log, email, rdx, cold
    :type bucket: str
    :param bucket: A GCS bucket name
    :rtype: str
    :return: A nicely formatted GCS path
    """
    segment = SEGMENTS.get(file_type)
    if not segment:
        msg = 'file_type is not one of these: {s}'
        raise ValueError(msg.format(s=', '.join(SEGMENTS)))
    if not bucket.startswith('gs://'):
        bucket = 'gs://{b}'.format(b=bucket)
    dname, bname = os.path.split(os.path.abspath(os.path.expanduser(fname)))
    if segment == 'COLD':
        gcs_file = bname
    else:
        gcs_file = '{d}/{f}'.format(
            d=os.path.basename(dname).replace('.', '_').replace('-', '_'),
            f=bname
        )
    return '{b}/{s}/{f}'.format(b=bucket, s=segment, f=gcs_file)


def course_to_bq_dataset(course_id: str, file_type: str, project: str) -> str:
    """
    Make a fully qualified BigQuery dataset name with the given info

    :type course_id: str
    :param course_id: edX course ID to format into a GCS path
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type project: str
    :param project: A GCP project ID
    :rtype: str
    :return: BigQuery dataset name with components separated by dots
    """
    if file_type not in SEGMENTS:
        msg = 'file_type is not one of these: {s}'
        raise ValueError(msg.format(s=', '.join(SEGMENTS)))
    if file_type in ('sql',):
        suffix = '_latest'
    elif file_type == 'log':
        suffix = '_logs'
    else:
        suffix = ''
    dataset = '{d}{s}'.format(
        d=course_id.replace('/', '__').replace('.', '_').replace('-', '_'),
        s=suffix
    )
    return '{p}.{d}'.format(p=project, d=dataset)


def local_to_bq_table(fname: str, file_type: str, project: str) -> str:
    """
    Use the given local file to make a fully qualified BigQuery table name

    :type fname: str
    :param fname: A local file name
    :type file_type: str
    :param file_type: One of sql, log, email, rdx
    :type project: str
    :param project: A GCP project ID
    :rtype: str
    :return: BigQuery dataset name with components separated by dots
    """
    if file_type not in SEGMENTS:
        raise ValueError(
            'file_type is not one of these: {s}'.format(
                s=', '.join(SEGMENTS)
            )
        )
    dname, bname = os.path.split(os.path.abspath(os.path.expanduser(fname)))
    if file_type in ('sql',):
        table = bname.split('.', 1)[0].replace('-', '_')
        suffix = '_latest'
    elif file_type in ('log',):
        suffix = '_logs'
        table = 'tracklog_' + ''.join(re.findall(r'\d+', bname))
    else:
        table = bname.split('.', 1)[0].replace('-', '_')
        suffix = ''
    dataset = os.path.basename(dname).replace('.', '_').replace('-', '_')
    if not table.replace('tracklog_', '') or not dataset:
        msg = (
            'A BigQuery table name for file type {t} '
            'could not be constructed with file name {f}. '
            'Please make sure that you are using the right file type.'
        )
        raise BigQueryNameException(
            msg.format(f=fname, t=file_type)
        )
    return '{p}.{d}{s}.{t}'.format(
        d=dataset, s=suffix, p=project, t=table
    )


def dict_to_schema_field(schema_dict: dict):
    """
    Make a SchemaField from a schema directory

    :type schema_dict: dict
    :param schema_dict: One of the objects in the schema JSON file
    :rtype: bigquery.SchemaField
    :returns: A SchemaField matching the given dictionary's name, type, etc.
    """
    if schema_dict.get('field_type') != 'RECORD':
        return bigquery.SchemaField(**schema_dict)
    fields = []
    for subfield in schema_dict.get('fields', []):
        fields.append(dict_to_schema_field(subfield))
    schema_dict['fields'] = fields
    field = bigquery.SchemaField(**schema_dict)
    return field


def get_bq_schema(table: str, schema_dir: str=SCHEMA_DIR):
    """
    Given a bare table name (without leading project or dataset),
    make a list of bigquery.SchemaField objects to act as the table's schema.

    :type table: str
    :param table: A BigQuery (bare) table name
    :type schema_dir: str
    :param schema_dir: Directory where schema JSON file is looked up
    :rtype: Tuple[List[bigquery.SchemaField], str]
    :return: A 2-tuple with list of bigquery.SchemaField objects and
        a description text for the target table
    :raises: MissingSchemaException
    """
    bname = table.split('.')[-1]
    if all(k in bname for k in ('track', 'log')):
        bname = 'tracking_log'
    targets = ('{t}.json', 'schema_{t}.json')
    for target in targets:
        sfile = os.path.join(schema_dir, target.format(t=bname))
        if not os.path.exists(sfile):
            continue
        out = []
        with open(sfile) as fh:
            schema = json.load(fh)
            for field in schema.get(bname, []):
                out.append(dict_to_schema_field(field))
        if not out:
            msg = (
                'The schema file {f!r} does not contain an object '
                'with a name matching the given table {t}'
            )
            raise MissingSchemaException(msg.format(f=sfile, t=table))
        return out, schema.get('description')
    msg = 'No JSON schema file found for {t} in directory {d}'
    raise MissingSchemaException(msg.format(t=table, d=schema_dir))


def make_bq_load_config(
    table: str, schema_dir=SCHEMA_DIR, append: bool=False,
    create: bool=True, file_format: str='json', delim=',', max_bad_rows=0,
):
    """
    Make a bigquery.LoadJobConfig object and description of a table

    :type table: str
    :param table: Fully qualified table name
    :type schema_dir: str
    :param schema_dir: The directory where schema files live
    :type append: bool
    :param append: Whether to append the loaded to the table
    :type create: bool
    :param create: Whether to create the target table if it does not exist
    :type file_format: str
    :param file_format: One of sql, json, csv, txt
    :type delim: str
    :param delim: The delimiter of the file being loaded
    :rtype: Tuple[bigquery.LoadJobConfig, str]
    :return: A 2-tuple with a bigquery.LoadJobConfig object and
        a description text for the destination table
    """
    schema_dir = schema_dir or SCHEMA_DIR
    schema, desc = get_bq_schema(table, schema_dir=schema_dir)
    if 'json' in file_format.lower():
        format_ = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
        delim = None
        skips = None
    else:
        format_ = bigquery.SourceFormat.CSV
        skips = 1
    if create:
        create = bigquery.CreateDisposition.CREATE_IF_NEEDED
    else:
        create = bigquery.CreateDisposition.CREATE_NEVER
    if append:
        append = bigquery.WriteDisposition.WRITE_APPEND
    else:
        append = bigquery.WriteDisposition.WRITE_TRUNCATE
    if 'json' in file_format.lower():
        config = bigquery.LoadJobConfig(
            schema=schema, source_format=format_,
            create_disposition=create, write_disposition=append,
            max_bad_records=max_bad_rows, ignore_unknown_values=True,
            destination_table_description=desc,
        )
        return config, desc
    config = bigquery.LoadJobConfig(
        schema=schema, source_format=format_,
        create_disposition=create, write_disposition=append,
        field_delimiter=delim, skip_leading_rows=skips,
        max_bad_records=max_bad_rows, ignore_unknown_values=True,
        allow_quoted_newlines=True, destination_table_description=desc,
    )
    return config, desc


def make_bq_query_config(append: bool=False, plain=True, table=None):
    """
    Make a bigquery.QueryJobConfig object to tie to a query to be sent
    to BigQuery for secondary table generation

    :type append: bool
    :param append: Whether to append the loaded to the table
    :type plain: bool
    :param plain: Make an empty QueryJobConfig object
    :type table: Union[None, str]
    :param table: Fully qualified name of a destination table
    :rtype: bigquery.QueryJobConfig
    :returns: Make a bigquery.QueryJobConfig object
    """
    if plain:
        return bigquery.job.QueryJobConfig()
    if append:
        append = bigquery.WriteDisposition.WRITE_APPEND
    else:
        append = bigquery.WriteDisposition.WRITE_TRUNCATE
    config = bigquery.job.QueryJobConfig()
    if table is not None:
        config.destination = bigquery.Table.from_string(table)
    config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
    config.write_disposition = append
    config.allow_large_results = True
    config.use_legacy_sql = False
    return config


def sqlify_bq_field(field, named=True):
    """
    Convert a bigquery.SchemaField object into a DDL
    column definition.

    :type field: bigquery.SchemaField
    :param field: A SchemaField object to convert to a DDL statement
    :type named: bool
    :param named: Whether the returned statement start with the field name
    :rtype: str
    :returns: A SQL column's DDL statement
    """
    nullability = '' if field.is_nullable else 'NOT NULL'
    if 'INTEGER' in field.field_type:
        type_ = 'INT64'
    elif 'FLOAT' in field.field_type:
        type_ = 'FLOAT64'
    else:
        type_ = field.field_type
    if type_ != 'RECORD':
        if field.mode != 'REPEATED':
            return '{n} {t} {m} OPTIONS(description="""{d}""")'.format(
                n=field.name if named else '',
                t=type_,
                m=nullability,
                d=field.description or '',
            )
        else:
            return '{n} ARRAY<{t}> {m} OPTIONS(description="""{d}""")'.format(
                n=field.name if named else '',
                t=type_,
                m=nullability,
                d=field.description or '',
            )
    if type_ == 'RECORD' and field.mode != 'REPEATED':
        return '{n} STRUCT<{t}> {m} OPTIONS(description="""{d}"""'.format(
            t=',\n\t'.join(sqlify_bq_field(f) for f in field.fields),
            n=field.name if named else '',
            m=nullability,
            d=field.description or '',
        )
    field = bigquery.SchemaField(
        name=field.name, field_type=field.field_type,
        mode='NULLABLE' if field.is_nullable else 'REQUIRED',
        description=field.description, fields=field.fields,
    )
    return '{n} ARRAY<{t}> {m} OPTIONS(description="""{d}""")'.format(
        t=sqlify_bq_field(field, False).lstrip(' '),
        n=field.name,
        m=nullability,
        d=field.description or '',
    )



"""
simeon-youtube is a companion script to the simeon tool that helps with
extracting YouTube video details like title and duration from the course_axis
files generated by ``simeon split`` with a sql file type.
"""
import glob
import gzip
import json
import os
import re
import sys
import traceback
import urllib.request as request
from argparse import (
    ArgumentParser, ArgumentTypeError,
    FileType, RawDescriptionHelpFormatter
)

import simeon
import simeon.scripts.utilities as cli_utils
import simeon.upload.gcp as gcp


SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)
API_URL = (
    'https://youtube.googleapis.com/youtube/v3/videos?'
    'part=contentDetails,snippet&id={ids}&key={token}'
)
VIDEO_COLS = {
    'id': ('id',), 'duration': ('contentDetails', 'duration'),
    'title': ('snippet', 'title'), 'description': ('snippet', 'description'),
    'channel_id': ('snippet', 'channelId'),
    'channel_title': ('snippet', 'channelTitle'),
    'published_at': ('snippet', 'publishedAt'),
}
DURATION_PATT = re.compile(r'^P(?P<date>\w*)T(?P<time>\w*)')
TIME_SECS = {
    'S': 1,
    'M': 60,
    'H': 3600,
}
DATE_SECS = {
    'D': 86400,
    'M': 2629800,
    'Y': 31557600,
}


def batch_size_type(val):
    """
    Make sure that val is an integer and is between 5 and 50.
    This is used for --batch-size to indicate the range of batch sizes
    allowed.
    """
    try:
        val = int(val)
    except Exception:
        raise ArgumentTypeError(
            '{v} is not a valid integer value. '
            'Please provide a valid number.'.format(v=val)
        )
    if val < 5 or val > 50:
        raise ArgumentTypeError('Expected batch sizes range between 5 and 50')
    return val


def _batch_ids(files, size=10):
    """
    Batch YouTube video IDs by the given size
    from the list of file names.
    """
    out = []
    seen = set()
    for file_ in files:
        for path in glob.iglob(file_):
            with gzip.open(path, 'rt') as fh:
                for line in fh:
                    try:
                        line = json.loads(line)
                    except Exception:
                        continue
                    id_ = (line.get('data') or {}).get('ytid', '')
                    if not id_:
                        continue
                    id_ = id_.split(':', 2)[-1].strip()
                    if id_ in seen:
                        continue
                    out.append(id_)
                    seen.add(id_)
                    if len(out) >= size:
                        yield out
                        out = []
    if out:
        yield out


def _generate_request(ids, token):
    """
    Create a urllib.request.Request object
    with the given ids and token values
    and send a GET request to the global API_URL.
    All the exceptions raised by urllib are propagated downstream
    """
    if isinstance(ids, (tuple, list, dict)):
        ids = ','.join(ids)
    url = API_URL.format(ids=ids, token=token)
    headers = {
        # 'Authorization': 'Bearer {t}'.format(t=token),
        'Accept': 'application/json',
    }
    req = request.Request(url, headers=headers)
    return request.urlopen(req)


def _process_one_by_one(ids, token, logger):
    """
    This steps through the video IDs and processes them one at a time
    """
    data = dict(items=[])
    for id_ in ids:
        try:
            items = json.load(_generate_request(id_, token)).get('items', [])
            data['items'] += items
        except Exception as excp:
            msg = 'Error processing video ID {i}: {e}'
            fh = getattr(excp, 'file', None)
            if fh is None:
                logger.error(msg.format(i=id_, e=excp))
            else:
                logger.error(msg.format(i=id_, e=json.load(fh)))
    return data


def _convert_and_time(val, times):
    """
    Convert the given value into an int and multiply by times
    """
    if not val:
        return 0
    try:
        return int(val) * times
    except Exception:
        return 0


def duration_to_seconds(duration):
    """
    Convert an ISO_8601 duration to seconds
    """
    durations = DURATION_PATT.search(duration)
    if not durations:
        return None
    out = 0
    names = {
        'time': TIME_SECS,
        'date': DATE_SECS,
    }
    for name, secs in names.items():
        chunk = durations.group(name)
        if not chunk:
            continue
        for char, times in secs.items():
            match = re.search(r'\d+(?={c})'.format(c=char), chunk)
            if not match:
                continue
            out += (_convert_and_time(match.group(0), times) or 0)
    return out


def extract_video_info(parsed_args):
    """
    Extract YouTube video details from the course axis files
    whose paths are provided via the given Namespace object.
    """
    keys = ('youtube-token',)
    seen = set()
    if not all(getattr(parsed_args, k.replace('-', '_'), None) for k in keys):
        msg = 'The following option(s) expected valid values: {o}'
        parsed_args.logger.error(msg.format(o=', '.join(keys)))
        sys.exit(1)
    os.makedirs(os.path.dirname(parsed_args.output), exist_ok=True)
    outfile = gzip.open(parsed_args.output, 'wt')
    for chunk in _batch_ids(parsed_args.course_axes, parsed_args.batch_size):
        data = None
        try:
            resp = _generate_request(chunk, parsed_args.youtube_token)
            data = json.load(resp)
        except Exception:
            data = _process_one_by_one(
                chunk, parsed_args.youtube_token, parsed_args.logger
            )
        if data is None:
            continue
        for item in data.get('items', []):
            record = {}
            for col, path in VIDEO_COLS.items():
                if len(path) == 1:
                    value = item.get(path[0])
                else:
                    subrec = (item.get(path[0]) or {})
                    for elm in path[1:-1]:
                        subrec = (subrec.get(elm) or {})
                    value = subrec.get(path[-1])
                record[col] = value
            if record.get('id') in seen:
                continue
            record['duration'] = duration_to_seconds(record['duration'])
            outfile.write(json.dumps(record) + '\n')
            seen.add(record.get('id'))
    outfile.close()


def merge_video_data(parsed_args):
    """
    Merge a data file to its target BigQuery table.
    This is a two-step process:
    1. Create a temporary table in BigQuery and load the data file
    2. Merge the temporary to the target table on a given column
    """
    keys = ('youtube-table', 'column', 'project')
    if not all(getattr(parsed_args, k.replace('-', '_'), None) for k in keys):
        msg = 'The following options expected valid values: {o}'
        parsed_args.logger.error(msg.format(o=', '.join(keys)))
        sys.exit(1)
    parsed_args.logger.info(
        'Merging {f} to {t}'.format(
            f=parsed_args.youtube_file, t=parsed_args.youtube_table
        )
    )
    parsed_args.logger.info('Connecting to BigQuery')
    try:
        if parsed_args.service_account_file is not None:
            client = gcp.BigqueryClient.from_service_account_json(
                parsed_args.service_account_file,
                project=parsed_args.project
            )
        else:
            client = client = gcp.BigqueryClient(
                project=parsed_args.project
            )
    except Exception as excp:
        errmsg = 'Failed to connect to BigQuery: {e}'
        parsed_args.logger.error(errmsg.format(e=excp))
        parsed_args.logger.error(
            'The error may be from an invalid service account file'
        )
        sys.exit(1)
    parsed_args.logger.info('Connection established')
    try:
        client.merge_to_table(
            fname=parsed_args.youtube_file, table=parsed_args.youtube_table,
            col=parsed_args.column,
            use_storage=parsed_args.youtube_file.startswith('gs://'),
            schema_dir=parsed_args.schema_dir,
            patch=parsed_args.update_description,
        )
    except Exception as excp:
        _, excp, tb = sys.exc_info()
        msg = 'Merging {f} to {t} failed with the following: {e}'
        if parsed_args.debug:
            traces = ['{e}'.format(e=excp)]
            traces += map(str.strip, traceback.format_tb(tb))
            msg = msg.format(
                e='\n'.join(traces),
                f=parsed_args.youtube_file,
                t=parsed_args.youtube_table,
            )
        else:
            msg = msg.format(
                e=excp,
                f=parsed_args.youtube_file,
                t=parsed_args.youtube_table,
            )
        parsed_args.logger.error(msg)
        sys.exit(1)
    msg = 'Successfully merged the records in {f} to the table {t}'
    parsed_args.logger.info(
        msg.format(
            f=parsed_args.youtube_file, t=parsed_args.youtube_table
        )
    )


def unknown_command(parsed_args):
    """
    Exit the program if an unknown command is passed (somehow)
    """
    parsed_args.logger.error(
        'Unknow command {c}'.format(c=parsed_args.command)
    )
    parsed_args.logger.error('Exiting...')
    sys.exit(1)


def main():
    """
    simeon-youtube entry point
    """
    parser = ArgumentParser(
        description=__doc__,
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '--log-file', '-L',
        help='Log file to use when simeon prints messages. Default: stdout',
        type=FileType('a'),
        default=sys.stdout,
    )
    parser.add_argument(
        '--debug', '-B',
        help='Show some stacktrace if simeon stops because of a fatal error',
        action='store_true',
    )
    parser.add_argument(
        '--quiet', '-Q',
        help='Only print error messages to standard streams.',
        action='store_false',
        dest='verbose',
    )
    parser.add_argument(
        '--config-file', '-C',
        help=(
            'The INI configuration file to use for default arguments.'
        ),
    )
    parser.add_argument(
        '--version', '-v',
        action='version',
        version='%(prog)s {v}'.format(v=simeon.__version__)
    )
    subparsers = parser.add_subparsers(
        description='A subcommand to carry out a task with simeon-youtube',
        dest='command'
    )
    subparsers.required = True
    extracter = subparsers.add_parser(
        'extract',
        help=(
            'Extract YouTube video details using the given API token and '
            'course axis json.gz files'
        ),
        description=(
            'Extract YouTube video details using the given API token and '
            'course axis json.gz files'
        )
    )
    extracter.add_argument(
        '--output', '-o',
        help=(
            'The .json.gz output file name for the YouTube video details. '
            'Default: %(default)s'
        ),
        default=os.path.join(os.getcwd(), 'youtube.json.gz'),
    )
    extracter.add_argument(
        '--youtube-token', '-t',
        help=(
            'YouTube data V3 API token generated from the account hosting '
            'the lecture videos.'
        )
    )
    extracter.add_argument(
        '--batch-size', '-s',
        help=(
            'How many YouTube video ID\'s to batch together when making HTTP '
            'requests for video metadata. Default: %(default)s'
        ),
        type=batch_size_type,
        default=10,
    )
    extracter.add_argument(
        'course_axes',
        help=(
            'course_axis.json.gz files from the simeon split process '
            'of SQL files'
        ),
        nargs='+',
    )
    merger = subparsers.add_parser(
        'merge',
        help=(
            'Merge the data file generated from simeon-youtube extract '
            'to the given target BigQuery table.'
        ),
        description=(
            'Merge the data file generated from simeon-youtube extract '
            'to the given target BigQuery table.'
        )
    )
    merger.add_argument(
        'youtube_file',
        help='A .json.gz file generated from the extract command'
    )
    merger.add_argument(
        '--project', '-p',
        help='The BigQuery project id where the target table resides.'
    )
    merger.add_argument(
        '--service-account-file', '-S',
        help='The service account file to use when connecting to BigQuery'
    )
    merger.add_argument(
        '--youtube-table', '-y',
        help='The target table where the YouTube video details are stored.',
        default='videos.youtube',
        type=cli_utils.bq_table,
    )
    merger.add_argument(
        '--column', '-c',
        help=(
            'The column on which to to merge the file and table. '
            'Default: %(default)s'
        ),
        default='id',
    )
    merger.add_argument(
        '--schema-dir', '-s',
        help=(
            'Directory where to find schema files. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    merger.add_argument(
        '--update-description', '-u',
        help=(
            'Update the description of the destination table with '
            'the "description" value from the corresponding schema file'
        ),
        action='store_true',
    )
    args = parser.parse_args()
    args.logger = cli_utils.make_logger(
        user='SIMEON-YOUTUBE:{c}'.format(c=args.command.upper()),
        verbose=args.verbose,
        stream=args.log_file,
    )
    try:
        configs = cli_utils.find_config(args.config_file)
    except Exception as excp:
        args.logger.error(str(excp).replace('\n', ' '))
        sys.exit(1)
    for k, v in cli_utils.CONFIGS.items():
        for (attr, cgetter) in v:
            cli_arg = getattr(args, attr, None)
            config_arg = cgetter(configs, k, attr, fallback=None)
            if not cli_arg and config_arg:
                setattr(args, attr, config_arg)
    COMMANDS = {
        'extract': extract_video_info,
        'merge': merge_video_data,
    }
    try:
        COMMANDS.get(args.command, unknown_command)(args)
    except:
        _, excp, tb = sys.exc_info()
        if isinstance(excp, SystemExit):
            raise excp
        msg = 'The command {c} failed: {e}'
        if args.debug:
            traces = ['{e}'.format(e=excp)]
            traces += map(str.strip, traceback.format_tb(tb))
            msg = msg.format(c=args.command, e='\n'.join(traces))
        else:
            msg = msg.format(c=args.command, e=excp)
        # msg = 'The command {c} failed with: {e}'
        args.logger.error(msg)
        sys.exit(1)


if __name__ == '__main__':
    main()




"""
Utility functions and classes to help with making course reports like user_info_combo, person_course, etc.
"""
import csv
import glob
import gzip
import json
import math
import multiprocessing as mp
import os
import re
import signal
import sys
import tarfile
import traceback
from collections import OrderedDict, defaultdict
from datetime import datetime
from functools import reduce
from multiprocessing.pool import (
    Pool as ProcessPool, ThreadPool, TimeoutError
)
from xml.etree import ElementTree

from dateutil.parser import parse as parse_date
from jinja2 import Template
from google.cloud.exceptions import NotFound

from simeon.download import utilities as downutils
from simeon.exceptions import (
    BadSQLFileException, EarlyExitError, LoadJobException,
    MissingFileException, MissingQueryFileException, MissingSchemaException,
    SchemaMismatchException, SQLQueryException,
)
from simeon.upload import gcp
from simeon.upload import utilities as uputils


# Increase the csv module's field size limit
csv.field_size_limit(13107200)

# BigQuery client for processes in process pools
report_bq_client = None


# Set up schema coercion functions
def _format_str_date(d):
    # return parse_date(d).strftime('%Y-%m-%d %H:%M:%S.%f')
    return parse_date(d).isoformat()


def _to_float(v):
    v = float(v)
    if math.isnan(v) or math.isinf(v):
        return None
    return v


def _stringify(v):
    if v is None:
        return None
    if not isinstance(v, str):
        v = json.dumps(v)
    return v.strip()


def _delete_incomplete_matches(dirname, patt):
    matches = glob.iglob(
        os.path.join(dirname, '*{p}*.json.gz'.format(p=patt))
    )
    for file_ in matches:
        try:
            os.remove(file_)
        except:
            continue


BQ2PY_TYPES = {
    'TIMESTAMP': _format_str_date,
    'STRING': _stringify,
    'INTEGER': int,
    'FLOAT': _to_float,
    'BOOLEAN': bool,
}


BQ_DDL = """#standardSQL
CREATE OR REPLACE TABLE {table} {cols}
OPTIONS (
    description = '''{description}'''
) AS
{query}"""
USER_INFO_COLS = OrderedDict([
    (
        ('auth_user-analytics.sql', None),
        [
            'user_id', 'username', 'email', 'is_staff',
            'last_login', 'date_joined',
        ],
    ),
    (
        ('auth_userprofile-analytics.sql', 'profile'),
        [
            'profile_name', 'profile_language', 'profile_location',
            'profile_meta', 'profile_courseware', 'profile_gender',
            'profile_mailing_address', 'profile_year_of_birth',
            'profile_level_of_education', 'profile_goals',
            'profile_allow_certificate', 'profile_country', 'profile_city',
        ],
    ),
    (
        ('student_courseenrollment-analytics.sql', 'enrollment'),
        [
            'enrollment_course_id', 'enrollment_created',
            'enrollment_is_active', 'enrollment_mode'
        ],
    ),
    (
        ('certificates_generatedcertificate-analytics.sql', 'certificate'),
        [
            'certificate_id', 'certificate_user_id',
            'certificate_download_url', 'certificate_grade',
            'certificate_course_id', 'certificate_key',
            'certificate_distinction', 'certificate_status',
            'certificate_verify_uuid', 'certificate_download_uuid',
            'certificate_name', 'certificate_created_date',
            'certificate_modified_date', 'certificate_error_reason',
            'certificate_mode'
        ],
    ),
    (
        ('user_id_map-analytics.sql', 'id_map'),
        ['id_map_hash_id']
    )
])
ADDED_COLS = [
    'edxinstructordash_Grade', 'edxinstructordash_Grade_timestamp',
    'y1_anomalous'
]
SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)
QUERY_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'queries',
)

PROBLEM_TYPES = {
    'choiceresponse', 'coderespons', 'customresponse',
    'fieldset', 'formularesponse', 'imageresponse',
    'multiplechoiceresponse', 'numericalresponse',
    'optionresponse', 'stringresponse', 'schematicresponse',
}


def _sql_pool_init():
    """
    Process pool initializer
    """
    def sighandler(sig, frame):
        raise EarlyExitError(
            'SQL table file generation interrupted prematurely'
        )
    sigs = [signal.SIGABRT, signal.SIGTERM, signal.SIGINT]
    for sig in sigs:
        signal.signal(sig, signal.SIG_DFL)


def _report_pool_init(proj, safile=None):
    """
    Process pool initializer
    """
    global report_bq_client
    def sighandler(sig, frame):
        raise EarlyExitError(
            'Secondary table generation interrupted prematurely'
        )
    sigs = [signal.SIGABRT, signal.SIGTERM, signal.SIGINT]
    for sig in sigs:
        signal.signal(sig, signal.SIG_DFL)
    if not safile:
        report_bq_client = gcp.BigqueryClient(project=proj)
    else:
        report_bq_client = gcp.BigqueryClient.from_service_account_json(
            safile, project=proj
        )


def _get_schema_dict(sdir, table):
    """
    Find a matching JSON schema file in the directory sdir
    for the given table
    """
    targets = ('schema_{t}.json', '{t}.json')
    for target in targets:
        schema_file = os.path.join(sdir, target.format(t=table))
        if os.path.exists(schema_file):
            with open(schema_file) as fh:
                return json.load(fh)
    msg = (
        'A schema file for table {t} could not be found in the given '
        'directory {d}'
    )
    raise MissingSchemaException(msg.format(t=table, d=sdir))


def wait_for_bq_jobs(job_list):
    """
    Given a list of BigQuery load or query jobs,
    wait for them all to finish.

    :type job_list: Iterable[LoadJob]
    :param job_list: An Iterable of job objects from the bigquery package
    :rtype: None
    :return: Nothing
    :TODO: Improve this function to behave a little less like a tight loop
    """
    done = set()
    while len(done) < len(job_list):
        for job in job_list:
            try:
                if job.job_id not in done:
                    state = job.done()
                    if state:
                        done.add(job.job_id)
            except NotFound:
                msg = '{id} is not a valid BigQuery job ID'
                raise LoadJobException(msg.format(id=job.job_id)) from None


def wait_for_bq_job_ids(job_list, client):
    """
    Given a list of BigQuery load or query job IDs,
    wait for them all to finish.

    :type job_list: Iterable[str]
    :param job_list: An Iterable of job IDs
    :rtype: Dict[str, Dict[str, str]]
    :return: Returns a dict of job IDs to job errors
    :TODO: Improve this function to behave a little less like a tight loop
    """
    out = dict()
    while len(out) < len(job_list):
        for job in job_list:
            if job not in out:
                try:
                    rjob = client.get_job(job)
                    if rjob.state == 'DONE':
                        src = ', '.join(getattr(rjob, 'source_uris', []))
                        err = (rjob.errors or [])
                        for e in err:
                            if e:
                                e['source'] = src
                        out[job] = err
                except NotFound:
                    msg = '{id} is not a valid BigQuery job ID'.format(id=job)
                    raise LoadJobException(msg) from None
    return out


def check_record_schema(record, schema, coerce=True, nullify=False):
    """
    Check that the given record matches the same keys found in the given
    schema list of fields. The latter is one of the schemas in
    simeon/upload/schemas/

    :type record: dict
    :param record: Dictionary whose values are modified
    :type schema: Iterable[Dict[str, Union[str, Dict]]]
    :param schema: A list of dicts with info on BigQuery table fields
    :type coerce: bool
    :param coerce: Whether or not to coerce values into BigQuery types
    :type nullify: bool
    :param nullify: Whether to set values mapping missing keys to None
    :rtype: None
    :returns: Modifies the record if needed
    :raises: SchemaMismatchException
    """
    for field in schema:
        if field.get('field_type') != 'RECORD':
            if field.get('name') not in record and nullify:
                if not coerce:
                    raise SchemaMismatchException(
                        '{f} is missing from the record'.format(
                            f=field.get('name')
                        )
                    )
                record[field.get('name')] = None
            elif field.get('name') in record and coerce:
                val = record[field.get('name')]
                func = BQ2PY_TYPES.get(field.get('field_type'))
                if func and val is not None:
                    try:
                        record[field.get('name')] = func(val)
                    except (ValueError, TypeError):
                        record[field.get('name')] = None
        else:
            subfields = field.get('fields')
            subrecord = record.get(field.get('name'), {})
            check_record_schema(subrecord, subfields, coerce)


def drop_extra_keys(record, schema):
    """
    Walk through the record and drop key-value pairs that are not in the
    given schema

    :type record: dict
    :param record: Dictionary whose values are modified
    :type schema: Iterable[Dict[str, Union[str, Dict]]]
    :param schema: A list of dicts with info on BigQuery table fields
    :rtype: None
    :return: Modifies the record if needed
    """
    if not schema:
        return
    keys = list(record)
    for k in keys:
        if k not in (f.get('name') for f in schema):
            del record[k]
        elif isinstance(record, dict) and isinstance(record.get(k), dict):
            subrecord = record.get(k, {})
            target = next((f for f in schema if f.get('name') == k), None)
            if target is None:
                subfields = []
            else:
                subfields = target.get('fields')
            drop_extra_keys(subrecord, subfields)


def extract_table_query(table, query_dir):
    """
    Given a table name and a query directory,
    extract both the query string and the table description.
    The latter is assumed to be any line in the query file
    that starts with # or --

    :type table: str
    :param table: BigQuery table name whose query info is extracted
    :type query_dir: str
    :param query_dir: The directory where the query file is expected to be
    :rtype: Tuple[str, str]
    :returns: A tuple of strings (query string, table description)
    :raises: MissingQueryFileException
    """
    table = table.split('.')[-1]
    qfile = os.path.join(query_dir, '{t}.sql'.format(t=table))
    if not os.path.exists(qfile):
        msg = 'Table {t} does not have a query file in {d}'
        raise MissingQueryFileException(msg.format(t=table, d=query_dir))
    with open(qfile) as qf:
        description = []
        query = []
        for line in qf:
            if line.startswith('--') or line.startswith('#'):
                description.append(line.lstrip('-# '))
            query.append(line)
    return (''.join(query), ''.join(description))


def make_user_info_combo(
    dirname, schema_dir=SCHEMA_DIR,
    outname='user_info_combo.json.gz'
):
    """
    Given a course's SQL directory, make a user_info_combo report

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the outname argument
    """
    schema_dir = schema_dir or SCHEMA_DIR
    for (file_, _) in USER_INFO_COLS:
        file_ = os.path.join(dirname, file_)
        if not os.path.exists(file_):
            raise OSError(
                '{f} does not exist in the SQL bundle.'.format(f=file_)
            )
    tbl = 'user_info_combo'
    schema = _get_schema_dict(schema_dir, tbl)[tbl]
    users = dict()
    user_file = 'auth_user-analytics.sql'
    user_cols = USER_INFO_COLS.get((user_file, None))
    with open(os.path.join(dirname, user_file)) as ufh:
        incols = [c.strip() for c in ufh.readline().split('\t')]
        reader = csv.DictReader(
            ufh, delimiter='\t', lineterminator='\n',
            quotechar='"', fieldnames=incols
        )
        for row in reader:
            uid = row.get('id')
            row['user_id'] = uid
            users[uid] = dict((k, row.get(k)) for k in user_cols)
    for (fname, prefix), cols in USER_INFO_COLS.items():
        if fname == user_file:
            continue
        with open(os.path.join(dirname, fname)) as rfh:
            header = []
            for col in map(str.strip, rfh.readline().split('\t')):
                if prefix:
                    uid_col = '{p}_user_id'.format(p=prefix)
                    username_col = '{p}_username'.format(p=prefix)
                    header.append('{p}_{c}'.format(p=prefix, c=col))
                else:
                    uid_col = 'user_id'
                    username_col = 'username'
                    header.append(col)
            reader = (dict(zip(header, r.split('\t'))) for r in rfh)
            # reader = csv.DictReader(
            #     (l.replace('"read more', 'read more') for l in rfh),
            #     delimiter='\t', lineterminator='\n',
            #     # quotechar='\'',
            #     fieldnames=header
            # )
            for row in reader:
                if uid_col not in row:
                    row[uid_col] = row.get('{p}_id'.format(p=prefix))
                user_id = row.get(uid_col)
                target = users.setdefault(user_id, {})
                target['user_id'] = user_id
                target.update(dict((k, row.get(k)) for k in cols))
                if target.get('username') is None and row.get(username_col):
                    target['username'] = row[username_col]
    outcols = reduce(lambda l, r: l + r, USER_INFO_COLS.values())
    outcols += ADDED_COLS
    with gzip.open(os.path.join(dirname, outname), 'wt') as zh:
        for record in users.values():
            outrow = dict()
            for k in outcols:
                val = record.get(k)
                val = val.strip() if val else val
                if 'course_id' in k:
                    val = downutils.get_sql_course_id(val or '') or None
                if 'certificate_grade' in k:
                    try:
                        val = str(float(val))
                    except (TypeError, ValueError):
                        val = None
                if val == 'NULL' or val == 'null':
                    outrow[k] = None
                else:
                    outrow[k] = val
            id_cols = ('user_id', 'certificate_user_id')
            if all(not outrow.get(k) for k in id_cols):
                continue
            # check_record_schema(outrow, schema, True)
            drop_extra_keys(outcols, schema)
            check_record_schema(outrow, schema)
            zh.write(json.dumps(outrow) + '\n')
        zh.flush()


def course_from_block(block):
    """
    Extract a course ID from the given block ID

    :type block: str
    :param block: A module item's block string
    :rtype: str
    :returns: Extracts the course ID in a module's block string
    """
    if block.startswith('i4x://'):
        return block.split('//')[-1].replace('course/', '')
    return '/'.join(block.split(':')[-1].split('+', 3)[:3])


def module_from_block(block):
    """
    Extract a module ID from the given block

    :type block: str
    :param block: A module item's block string
    :rtype: str
    :returns: Extracts the module ID in a module's block string
    """
    block = block.replace('/courses/course-v1:', '')
    if block.startswith('i4x://'):
        return block.lstrip('i4x://')
    segments = block.split(':', 1)[-1].split('+')
    segments = '/'.join(map(lambda s: s.split('@')[-1], segments))
    return '/'.join(segments.split('/')[:5])


def get_youtube_id(record):
    """
    Given a course structure record, extract the YouTube ID
    associated with the video element.

    :type record: dict
    :param record: A course_axis record
    :rtype: Union[str, None]
    :returns: The YouTube video ID associated with the record
    """
    for k, v in record.get('metadata', {}).items():
        if 'youtube_id' in k and v:
            return ':'.join(re.findall(r'\d+', k) + [v])


def _get_itypes(fname, problem_types=PROBLEM_TYPES):
    """
    Extract values for the course_axis.data.itype field
    from the given tar file.
    """
    out = dict()
    with tarfile.open(fname) as tf:
        for problem in tf.getmembers():
            if '/problem/' not in problem.name or problem.isdir():
                continue
            block = os.path.splitext(problem.name)[0].split('/')[-1]
            pf = tf.extractfile(problem)
            root = ElementTree.fromstring(pf.read())
            for elm in root:
                if elm.tag in problem_types:
                    out[block] = elm.tag
                    break
    return out


def get_has_solution(record):
    """
    Extract whether the given record is a problem that has showanswer.
    If it's present and its associated value is not "never", then return True.
    Otherwise, return False.

    :type record: dict
    :param record: A course_axis record
    :rtype: bool
    :returns: Wether the course_axis record has a solution in the data
    """
    meta = record.get('metadata') or dict()
    if 'showanswer' not in meta:
        return False
    return meta['showanswer'] != 'never'


def get_problem_nitems(record):
    """
    Get a value for data.num_items in course_axis

    :type record: dict
    :param record: A course_axis record
    :rtype: Union[int, None]
    :returns: The number of subitems of a problem item
    """
    if 'problem' in record.get('category', ''):
        return len(record.get('children', [])) + 1
    return None


def _get_axis_path(block, mapping):
    """
    Extract the path to the given block from the root
    of the course structure.

    :type block: str
    :param block: One of the block IDs or names from the course structure file
    :type mapping: dict
    :param mapping: A dict mapping child blocks to their parents
    :rtype: str
    :return: A constructed path from root to the given block with hashes
    """
    if '/course/' in block:
        return '/'
    path = []
    while block:
        path.append(module_from_block(block))
        block = mapping.get(block)
    return '/{p}'.format(p='/'.join(
        map(lambda s: s.split('/')[-1], path[-2::-1])
    ))


def _get_first_axis_meta(block, name, struct, mapping):
    """
    Get the first non-null or non-empty value
    of the given metadata name from the data dictionary
    starting at the given block.
    Use the mapping dictionary to find the parents of items
    to consider
    """
    struct = (struct or dict())
    out = None
    while block:
        out = struct.get(block, {}).get('metadata', {}).get(name)
        if out:
            break
        block = mapping.get(block)
    return out


def process_course_structure(data, start, mapping, parent=None):
    """
    The course structure data dictionary and starting point,
    loop through it and construct course axis data items

    :type data: dict
    :param data: The data from the course_structure-analytics.json file
    :type start: str
    :param start: The key from data to start looking up children
    :type mapping: dict
    :param mapping: A dict mapping child blocks to their parents
    :type parent: Union[None, str]
    :param parent: Parent of start
    :rtype: List[Dict]
    :return: Returns the list of constructed data items
    """
    out = []
    record = data.get(start, {})
    sep = '/' if start.startswith('i4x:') else '@'
    children = record.get('children', [])
    item = dict(
        parent=parent.split(sep)[-1] if parent else None,
        split_url_name=None,
    )
    item['path'] = _get_axis_path((start or '/course/'), mapping)
    item['category'] = record.get('category', '')
    item['url_name'] = start.split(sep)[-1]
    targets = (
        ('name', 'display_name'), ('gformat', 'format'),
        ('due', 'due'), ('start', 'start'), ('graded', 'graded'),
        ('visible_to_staff_only', 'visible_to_staff_only'),
    )
    for (key, target) in targets:
        item[key] = _get_first_axis_meta(
            block=start, name=target,
            struct=data, mapping=mapping
        )
    item['graded'] = bool(item.get('graded'))
    item['is_split'] = any([
        'split_test' in item['category'],
        'split_test' in data.get(parent, {}).get('category', '')
    ])
    if item['is_split']:
        if 'split_test' in start:
            item['split_url_name'] = item['url_name']
        else:
            item['split_url_name'] = item['parent']
    item['module_id'] = module_from_block(start)
    item['data'] = dict(
        ytid=get_youtube_id(record),
        weight=record.get('metadata', {}).get('weight'),
        group_id_to_child=None,
        user_partition_id=None,
        itype=None,
        num_items=get_problem_nitems(record),
        has_solution=get_has_solution(record),
        has_image=False,
    )
    out.append(item)
    if children:
        for child in children:
            out.extend(
                process_course_structure(
                    data=data, start=child,
                    parent=start, mapping=mapping,
                )
            )
    return out


def make_course_axis(
    dirname, schema_dir=SCHEMA_DIR, outname='course_axis.json.gz'
):
    """
    Given a course's SQL directory, make a course_axis report

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the outname argument
    """
    schema_dir = schema_dir or SCHEMA_DIR
    fname = os.path.join(dirname, 'course_structure-analytics.json')
    bundle = os.path.join(dirname, 'course-analytics.xml.tar.gz')
    for file_ in (fname, bundle):
        if not os.path.exists(file_):
            raise OSError(
                '{f} does not exist in the SQL bundle.'.format(f=file_)
            )
    itypes = _get_itypes(bundle)
    with open(fname) as fh:
        structure: dict = json.load(fh)
    # Find the course object (i.e. root object)
    root_block = None
    root_val = None
    for block, val in structure.items():
        if val.get('category') == 'course':
            root_block = block
            root_val = val
            break
    if not root_block:
        msg = (
            'The given course structure file {f!r} does not have a root'
            ' course block. Please reach out to edX to have them fix it.'
        )
        raise BadSQLFileException(msg.format(f=fname))
    course_id = course_from_block(root_block)
    # Map child items to their parents
    child_to_parent = dict()
    for k, v in structure.items():
        children = v.get('children') or []
        for child in children:
            child_to_parent[child] = k
    data = process_course_structure(
        data=structure, start=root_block,
        mapping=child_to_parent,
    )
    outname = os.path.join(dirname, 'course_axis.json.gz')
    with gzip.open(outname, 'wt') as zh:
        chapter_mid = None
        for index, record in enumerate(data, 1):
            if record.get('category') == 'chapter':
                chapter_mid = record.get('module_id')
            record['course_id'] = course_id
            record['chapter_mid'] = chapter_mid
            record['index'] = index
            record['data']['itype'] = itypes.get(
                record.get('module_id', '').split('/')[-1]
            )
            if record['gformat']:
                if not record.get('due'):
                    record['due'] = root_val.get('end')
                if not record.get('start'):
                    record['start'] = root_val.get('start')
            zh.write(json.dumps(record) + '\n')
        zh.flush()


def make_grades_persistent(
    dirname,
    schema_dir=SCHEMA_DIR,
    first_outname='grades_persistent.json.gz',
    second_outname='grades_persistent_subsection.json.gz'
):
    """
    Given a course's SQL directory, make the grades_persistent
    and grades_persistent_subsection reports.

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the target files
    """
    schema_dir = schema_dir or SCHEMA_DIR
    infiles = dict([
        (
            'grades_persistentcoursegrade-analytics.sql',
            (first_outname, 'grades_persistent'),
        ),
        (
            'grades_persistentsubsectiongrade-analytics.sql',
            (second_outname, 'grades_persistent_subsection'),
        )
    ])
    for file_, (outname, tbl) in infiles.items():
        file_ = os.path.join(dirname, file_)
        outname = os.path.join(dirname, outname)
        if not os.path.exists(file_):
            raise OSError(
                '{f} does not exist in the SQL bundle.'.format(f=file_)
            )
        schema = _get_schema_dict(schema_dir, tbl).get(tbl)
        with open(file_) as gh, gzip.open(outname, 'wt') as zh:
            header = [c.strip() for c in gh.readline().split('\t')]
            reader = csv.DictReader(
                gh, delimiter='\t', quotechar='\'',
                lineterminator='\n', fieldnames=header
            )
            for record in reader:
                for k in record:
                    if 'course_id' in k:
                        record[k] = downutils.get_sql_course_id(record[k])
                    if record[k] == 'NULL':
                        record[k] = None
                check_record_schema(record, schema)
                zh.write(json.dumps(record) + '\n')
            zh.flush()


def make_grading_policy(
    dirname,
    schema_dir=SCHEMA_DIR,
    outname='grading_policy.json.gz'
):
    """
    Generate a file to be loaded into the grading_policy table
    of the given SQL directory.

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the target file
    """
    schema_dir = schema_dir or SCHEMA_DIR
    file_ = os.path.join(dirname, 'course-analytics.xml.tar.gz')
    if not os.path.exists(file_):
        raise OSError(
            '{f} does not exist in the SQL bundle'.format(f=file_)
        )
    with tarfile.open(file_) as tar:
        policy = next(
            (m for m in tar.getmembers() if 'grading_policy.json' in m.name),
            None
        )
        if policy is None:
            raise MissingFileException(
                'No grading policy found in {f!r}'.format(f=file_)
            )
        with tar.extractfile(policy) as jh:
            grading_policy = json.load(jh)
        outname = os.path.join(dirname, outname)
        cols = (
            'assignment_type', 'name', 'fraction_of_overall_grade',
            'min_count', 'drop_count', 'short_label',
            'overall_lower_cutoff', 'overall_lower_cutoff_label',
            'overall_upper_cutoff', 'overall_upper_cutoff_label',
        )
        with gzip.open(outname, 'wt') as zh:
            for grader in grading_policy.get('GRADER', []):
                grader['assignment_type'] = grader.get('type')
                grader['name'] = grader.get('type')
                grader['fraction_of_overall_grade'] = grader.get('weight')
                cutoffs = sorted(
                    (grading_policy.get('GRADE_CUTOFFS') or {}).items(),
                    key=lambda t: t[1]
                )
                if cutoffs:
                    grader['overall_lower_cutoff'] = cutoffs[0][1]
                    grader['overall_lower_cutoff_label'] = cutoffs[0][0]
                    grader['overall_upper_cutoff'] = cutoffs[-1][1]
                    grader['overall_upper_cutoff_label'] = cutoffs[-1][0]
                else:
                    grader['overall_lower_cutoff'] = None
                    grader['overall_lower_cutoff_label'] = None
                    grader['overall_upper_cutoff'] = None
                    grader['overall_upper_cutoff_label'] = None
                zh.write(
                    json.dumps(dict((k, grader.get(k)) for k in cols)) + '\n'
                )
            zh.flush()


def _extract_mongo_values(record, key, subkey):
    """
    Given a forum data dictionary with immediate and sub keys,
    extract the value(s) at subkey and jsonify them, if need be.
    """
    target = record.get(key)
    if not target:
        return target
    if isinstance(target, dict):
        return target.get(subkey)
    if isinstance(target, list):
        out = []
        for subrec in target:
            if subkey in subrec:
                out.append(subrec[subkey])
        return json.dumps(out)
    return None


def make_forum_table(
    dirname,
    schema_dir=SCHEMA_DIR,
    outname='forum.json.gz'
):
    """
    Generate a file to load into the forum table
    using the given SQL directory

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type schema_dir: str
    :param schema_dir: Directory where schema files live
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the target file
    """
    schema_dir = schema_dir or SCHEMA_DIR
    outname = os.path.join(dirname, outname)
    file_ = os.path.join(dirname, 'forum.mongo')
    if not os.path.exists(file_):
        raise OSError(
            '{f} does not exist in the SQL bundle.'.format(f=file_)
        )
    cols = {
        '$oid': (
            '_id', 'parent_id', 'parent_ids', 'comment_thread_id',
        ),
        '$date': (
            ('endorsement', 'time'), 'updated_at',
            'created_at', 'last_activity_at'
        ),
        None: (
            '_type', 'abuse_flaggers', 'anonymous', 'anonymous_to_peers',
            'at_position_list', 'author_id', 'author_username', 'body',
            'child_count', 'closed', 'comment_count', 'commentable_id',
            'context', 'course_id', 'depth', 'endorsed',
            'historical_abuse_flaggers', 'pinned', 'retired_username',
            'sk', 'thread_type', 'title', 'visible', 'votes'
        ),
    }
    schema = _get_schema_dict(schema_dir, 'forum')['forum']
    with open(file_) as fh, gzip.open(outname, 'wt') as zh:
        for line in fh:
            record = json.loads(line)
            for subkey, keys in cols.items():
                for col in keys:
                    if isinstance(col, (tuple, list)):
                        col, subcol = col[:2]
                    else:
                        subcol = None
                    if subkey is not None:
                        if subcol:
                            val = _extract_mongo_values(
                                (record.get(col, {}) or {}), subcol, subkey
                            )
                            if not isinstance(record.get(col), dict):
                                record[col] = {}
                            record[col][subcol] = val
                        else:
                            record[col] = _extract_mongo_values(
                                record, col, subkey
                            )
                    if record.get(col, '') == 'NULL':
                        record[col] = None
                    if isinstance(record.get(col), list):
                        record[col] = json.dumps(record[col])
                    if isinstance(record.get(col), dict):
                        for k in record[col]:
                            if isinstance(record[col][k], list):
                                record[col][k] = json.dumps(record[col][k])
            record['mongoid'] = record['_id']
            record['course_id'] = downutils.get_sql_course_id(
                record.get('course_id') or ''
            )
            drop_extra_keys(record, schema)
            check_record_schema(record, schema, True)
            zh.write(json.dumps(record) + '\n')
        zh.flush()


def make_problem_analysis(state, **extras):
    """
    Use the state record from a studentmodule record to make
    a record to load into the problem_analysis table.
    The state is assumed to be from record of category "problem".

    :type state: dict
    :param state: Contents of the state field of studentmodule
    :type extras: keyword arguments
    :param extras: Things to be added to the generated record
    :rtype: dict
    :return: Return a record to be loaded in problem_analysis
    """
    maps = state.get('correct_map') or {}
    answers = state.get('student_answers') or {}
    items = []
    for k, v in maps.items():
        items.append(
            {
                'answer_id': k,
                'correctness': v.get('correctness'),
                'correct_bool' : v.get('correctness', '') == 'correct',
                'npoints': v.get('npoints'),
                'msg': v.get('msg'),
                'hint': v.get('hint'),
                'response': json.dumps(answers.get(k)),
            }
        )
    out = {
        'item': items,
        'attempts': state.get('attempts', 0),
        'done': state.get('done'),
    }
    out.update(extras)
    return out


def make_student_module(
    dirname,
    schema_dir=SCHEMA_DIR,
    outname='studentmodule.json.gz'
):
    """
    Generate files to load into studentmodule and problem_analysis
    using the given SQL directory

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type schema_dir: str
    :param schema_dir: Directory where schema files live
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the target files
    """
    schema_dir = schema_dir or SCHEMA_DIR
    pjoin = os.path.join
    outname = pjoin(dirname, outname)
    second = pjoin(dirname, 'problem_analysis.json.gz')
    file_ = pjoin(dirname, 'courseware_studentmodule-analytics.sql')
    if not os.path.exists(file_):
        raise OSError(
            '{f} does not exist in the SQL bundle.'.format(f=file_)
        )
    tbl = 'studentmodule'
    module_schema = _get_schema_dict(schema_dir, tbl).get(tbl)
    tbl = 'problem_analysis'
    problem_schema = _get_schema_dict(schema_dir, tbl).get(tbl)
    prob_cols = ('correct_map', 'student_answers')
    with open(file_, encoding='UTF8', errors='ignore') as fh:
        header = [c.strip() for c in fh.readline().split('\t')]
        reader = csv.DictReader(
            (l.replace('\0', '') for l in fh),
            delimiter='\t', quotechar='\'', lineterminator='\n',
            fieldnames=header
        )
        with gzip.open(outname, 'wt') as zh, gzip.open(second, 'wt') as ph:
            for record in reader:
                for k, v in record.items():
                    if k == 'course_id':
                        record[k] = downutils.get_sql_course_id(v or '')
                    if k == 'module_id':
                        record[k] = module_from_block(v or '')
                    if (v or '').lower() == 'null':
                        record[k] = None
                check_record_schema(record, module_schema)
                zh.write(json.dumps(record) + '\n')
                try:
                    state = json.loads(
                        (record.get('state') or '{}').replace('\\\\', '\\')
                    )
                except json.JSONDecodeError:
                    continue
                if not all(k in state for k in prob_cols):
                    continue
                url_name = record.get('module_id', '').split('/')[-1]
                panalysis = make_problem_analysis(
                    state, course_id=record.get('course_id'),
                    user_id=record.get('student_id'),
                    problem_url_name=url_name,
                    grade=record.get('grade', 0),
                    max_grade=record.get('max_grade', 0),
                    created=record.get('created')
                )
                check_record_schema(panalysis, problem_schema)
                ph.write(json.dumps(panalysis) + '\n')
            zh.flush()
            ph.flush()


def _default_roles():
    """
    Generate a default roles record to give to a defaultdict
    """
    return {
        'course_id': None, 'user_id': None, 'roles_isBetaTester': 0,
        'roles_isInstructor': 0, 'roles_isStaff': 0,
        'roles_isCCX': 0, 'roles_isFinance': 0,
        'roles_isLibrary': 0, 'roles_isSales': 0,
        'forumRoles_isAdmin': 0, 'forumRoles_isCommunityTA': 0,
        'forumRoles_isModerator': 0, 'forumRoles_isStudent': 0,
        'roles': None,
    }


def make_roles_table(
    dirname,
    schema_dir=SCHEMA_DIR,
    outname='roles.json.gz'
):
    """
    Generate a file to be loaded into the roles table of a dataset

    :type dirname: str
    :param dirname: Name of a course's directory of SQL files
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :type outname: str
    :param outname: The filename to give it to the generated report
    :rtype: None
    :return: Nothing, but writes the generated data to the target files
    """
    schema_dir = schema_dir or SCHEMA_DIR
    files = {
        'student_courseaccessrole-analytics.sql',
        'django_comment_client_role_users-analytics.sql',
    }
    roles = {
        'beta_testers': 'roles_isBetaTester',
        'ccx_coach': 'roles_isCCX',
        'finance_admin': 'roles_isFinance',
        'instructor': 'roles_isInstructor',
        'library_user': 'roles_isLibrary',
        'sales_admin': 'roles_isSales',
        'staff': 'roles_isStaff',
        'Administrator': 'forumRoles_isAdmin',
        'Community': 'forumRoles_isCommunityTA',
        'Community TA': 'forumRoles_isCommunityTA',
        'Moderator': 'forumRoles_isModerator',
        'Student': 'forumRoles_isStudent',
    }
    with gzip.open(os.path.join(dirname, outname), 'wt') as zh:
        data = defaultdict(_default_roles)
        for file_ in map(lambda f: os.path.join(dirname, f), files):
            if not os.path.exists(file_):
                raise OSError(
                    '{f} does not exist in the SQL bundle.'.format(f=file_)
                )
            with open(file_) as fh:
                line = fh.readline().replace('\tname', '\trole')
                header = []
                for c in line.split('\t'):
                    header.append(c.strip())
                reader = csv.DictReader(
                    fh, delimiter='\t', quotechar='\'', lineterminator='\n',
                    fieldnames=header
                )
                for inrow in reader:
                    outrow = data[inrow.get('user_id')]
                    outrow['user_id'] = inrow.get('user_id')
                    outrow['course_id'] = downutils.get_sql_course_id(
                        inrow.get('course_id', '')
                    )
                    col = roles.get(inrow.get('role'))
                    if col is not None:
                        outrow[col] = 1
                    if 'Student' in inrow.get('role', ''):
                        rval = 'Student'
                    else:
                        rval = 'Staff'
                    outrow['roles'] = rval
        staff = set(k for k in roles.values() if k.startswith('roles_'))
        for record in data.values():
            if any(record.get(k) for k in staff):
                record['roles'] = 'Staff'
            zh.write(json.dumps(record) + '\n')
        zh.flush()


def make_sql_tables_seq(
    dirnames, verbose=False, logger=None, fail_fast=False, debug=False,
    schema_dir=SCHEMA_DIR,
):
    """
    Given an iterable of SQL directories, make the SQL tables
    defined in this module.
    This convenience function calls all the report generating functions
    for the given directory name

    :type dirnames: Iterable[str]
    :param dirnames: Names of SQL directories
    :type verbose: bool
    :param verbose: Print a message when a report is being made
    :type logger: logging.Logger
    :param logger: A logging.Logger object to print messages with
    :type fail_fast: bool
    :param fail_fast: Whether or not to bail after the first error
    :type debug: bool
    :param debug: Show the stacktrace that caused the error
    :type schema_dir: str
    :param schema_dir: The directory where schema files live
    :rtype: bool
    :return: True if the files are generated, and False otherwise.
    """
    schema_dir = schema_dir or SCHEMA_DIR
    reports = (
        make_course_axis, make_forum_table, make_grades_persistent,
        make_grading_policy, make_roles_table,
        make_student_module, make_user_info_combo,
    )
    fails = []
    for dirname in dirnames:
        for fn in reports:
            tbl = fn.__name__.replace('make_', '').replace('_table', '')
            if verbose and logger is not None:
                msg = 'Making {f} with files in {d}'
                logger.info(msg.format(f=tbl, d=dirname))
            try:
                fn(dirname=dirname, schema_dir=schema_dir)
            except:
                _, excp, tb = sys.exc_info()
                if debug:
                    traces = ['{e}'.format(e=excp)]
                    traces += map(str.strip, traceback.format_tb(tb))
                    excp = '\n'.join(traces)
                msg = (
                    'Error encountered while making the {n} table(s) '
                    'with the given directory {d}: {e}'
                )
                excp = MissingFileException(msg.format(
                    d=dirname, n=tbl, e=excp
                ))
                if fail_fast:
                    raise excp from None
                fails.append(excp)
                _delete_incomplete_matches(dirname, tbl)
        if verbose and logger is not None:
            for failure in fails:
                logger.error(failure)
            logger.info('Done processing files in {d}'.format(d=dirname))
    return not bool(fails)


def make_sql_tables_par(
    dirnames, verbose=False, logger=None, fail_fast=False, debug=False,
    schema_dir=SCHEMA_DIR
):
    """
    Given a list of SQL directories, make the SQL tables
    defined in this module.
    This convenience function calls all the report generating functions
    for the given directory name

    :type dirnames: List[str]
    :param dirnames: Names of SQL directories
    :type verbose: bool
    :param verbose: Print a message when a report is being made
    :type logger: logging.Logger
    :param logger: A logging.Logger object to print messages with
    :type fail_fast: bool
    :param fail_fast: Whether or not to bail after the first error
    :type debug: bool
    :param debug: Show the stacktrace that caused the error
    :type schema_dir: str
    :param schema_dir: The directory where schema files live
    :rtype: bool
    :return: True if the files are generated, and False otherwise.
    """
    schema_dir = schema_dir or SCHEMA_DIR
    reports = (
        make_course_axis, make_forum_table, make_grades_persistent,
        make_grading_policy, make_roles_table,
        make_student_module, make_user_info_combo,
    )
    results = dict()
    nprocs = mp.cpu_count()
    if len(dirnames) < nprocs:
        nprocs = len(dirnames)
    with ProcessPool(nprocs, initializer=_sql_pool_init) as pool:
        for fn in reports:
            for dirname in dirnames:
                tbl = fn.__name__.replace('make_', '').replace('_table', '')
                if verbose and logger is not None:
                    msg = 'Making {f} with files in {d}'
                    logger.info(msg.format(f=tbl, d=dirname))
                results[(tbl, dirname)] = pool.apply_async(
                    fn, args=(dirname, schema_dir)
                )
        fails = []
        for (tbl, dirname), result in results.items():
            try:
                result.get()
            except KeyboardInterrupt:
                logger.error(
                    'Report generation interrupted by user.'
                )
                raise EarlyExitError()
            except:
                _, excp, tb = sys.exc_info()
                if debug:
                    traces = ['{e}'.format(e=excp)]
                    traces += map(str.strip, traceback.format_tb(tb))
                    excp = '\n'.join(traces)
                msg = (
                    'Error encountered while making the {n} table(s) '
                    'with the given directory {d}: {e}'
                )
                excp = MissingFileException(msg.format(
                    d=dirname, n=tbl, e=excp
                ))
                if fail_fast:
                    raise excp from None
                fails.append(excp)
                _delete_incomplete_matches(dirname, tbl)
    if verbose and logger is not None:
        for failure in fails:
            logger.error(failure)
        for dirname in dirnames:
            logger.info('Done processing files in {d}'.format(d=dirname))
    return not bool(fails)


def make_table_from_sql(
    table, course_id, client, project, append=False,
    query_dir=QUERY_DIR, schema_dir=SCHEMA_DIR,
    wait=False, geo_table='geocode.geoip', youtube_table='videos.youtube',
    **kwargs,
):
    """
    Generate a BigQuery table using the given table name,
    course ID and a matching SQL query file in the query_dir folder.
    The query file contains placeholder for course ID, dataset name and
    other details.

    :type table: str
    :param table: table name
    :type course_id: str
    :param course_id: Course ID whose secondary reports are being generated
    :type client: bigquery.Client
    :param client: An authenticated bigquery.Client object
    :type project: str
    :param project: GCP project id where the video_axis table is loaded.
    :type query_dir: Union[None, str]
    :param query_dir: Directory where query files are saved.
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :type geo_table: str
    :param geo_table: Table name in BigQuery with geolocation data for IPs
    :type youtube_table: str
    :param youtube_table: Table name in BigQuery with YouTube video details
    :type wait: bool
    :param wait: Whether to wait for the query job to finish running
    :rtype: Dict[str, Dict[str, str]]
    :return: Returns the errors dictionary from the LoadJob object tied to the query
    """
    schema_dir = schema_dir or SCHEMA_DIR
    query_dir = query_dir or QUERY_DIR
    # Don't get extra_args overwrite the given course ID
    kwargs.pop('course_id', None)
    kwargs.pop('latest_dataset', None)
    kwargs.pop('log_dataset', None)
    if course_id != kwargs.get('course_id'):
        kwargs.pop('course_id', None)
    latest_dataset = uputils.course_to_bq_dataset(
        course_id, 'sql', project
    )
    log_dataset = uputils.course_to_bq_dataset(
        course_id, 'log', project
    )
    try:
        cols = []
        fields, schema_desc = uputils.get_bq_schema(table, schema_dir)
        for f in fields:
            cols.append(uputils.sqlify_bq_field(f))
        cols = ',\n'.join(cols)
    except MissingSchemaException:
        cols = ''
        schema_desc = ''
    query, description = extract_table_query(table, query_dir)
    table = '{d}.{t}'.format(d=latest_dataset, t=table)
    if append:
        config = uputils.make_bq_query_config(
            append=True, plain=False, table=table
        )
        config.destination = table
    else:
        config = uputils.make_bq_query_config(plain=True)
        query = BQ_DDL.format(
            table=table,
            description=(schema_desc or description.strip()),
            query=query,
            cols='({c})'.format(c=cols) if cols else '',
        )
    query = Template(query).render(
        geo_table=geo_table, youtube_table=youtube_table,
        course_id=course_id, latest_dataset=latest_dataset,
        log_dataset=log_dataset, **kwargs
    )
    try:
        job = client.query(
            query.format(
                latest_dataset=latest_dataset, log_dataset=log_dataset,
                geo_table=geo_table, youtube_table=youtube_table,
                course_id=course_id, **kwargs
            ),
            job_id_prefix='{t}_{dt}'.format(
                t=table.replace('.', '_'),
                dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
            ),
            job_config=config,
        )
    except Exception as excp:
        errors = getattr(excp, 'errors', [None])
        if errors and errors[0]:
            msg = errors[0].get('message')
        else:
            msg = str(excp)
        raise SQLQueryException(msg)
    if wait:
        status = wait_for_bq_job_ids([job.job_id], client)
        return status[job.job_id]
    return job.errors or {}


def make_tables_from_sql(
    tables, course_id, client, project, append=False,
    query_dir=QUERY_DIR, wait=False,
    geo_table='geocode.geoip', youtube_table='videos.youtube',
    parallel=False, fail_fast=False,
    schema_dir=SCHEMA_DIR, **kwargs
):
    """
    This is the plural/multiple tables version of make_table_from_sql

    :type tables: Iterable[str]
    :param tables: BigQuery table names to create or append to
    :type course_id: str
    :param course_id: Course ID whose secondary reports are being generated
    :type client: bigquery.Client
    :param client: An authenticated bigquery.Client object
    :type project: str
    :param project: GCP project id where the video_axis table is loaded.
    :type query_dir: Union[None, str]
    :param query_dir: Directory where query files are saved.
    :type geo_table: str
    :param geo_table: Table name in BigQuery with geolocation data for IPs
    :type youtube_table: str
    :param youtube_table: Table name in BigQuery with YouTube video details
    :type wait: bool
    :param wait: Whether to wait for the query job to finish running
    :type parallel: bool
    :param parallel: Whether the function is running in a process pool
    :type fail_fast: bool
    :param fail_fast: Whether to stop processing after the first error
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :rtype: Dict[str, Dict[str, str]]
    :return: Return a dict mapping table names to their corresponding errors
    """
    query_dir = query_dir or QUERY_DIR
    schema_dir = schema_dir or SCHEMA_DIR
    # Don't get extra_args overwrite the given course ID
    kwargs.pop('course_id', None)
    if parallel:
        global report_bq_client
        client = report_bq_client
    out = dict()
    dataset = '{p}.{c}_latest'.format(
        p=project, c=course_id.replace('/', '__').replace('.', '_')
    )
    client.create_dataset(dataset, exists_ok=True)
    for table in tables:
        out[table] = make_table_from_sql(
            table=table, course_id=course_id, client=client, project=project,
            append=append, geo_table=geo_table, wait=wait,
            query_dir=query_dir, youtube_table=youtube_table,
            schema_dir=schema_dir, **kwargs
        )
        if fail_fast and out[table]:
            return out
    return out


def make_tables_from_sql_par(
    tables, courses, project, append=False, query_dir=QUERY_DIR,
    wait=False, geo_table='geocode.geoip', youtube_table='videos.youtube',
    safile=None, size=mp.cpu_count(), logger=None, fail_fast=False,
    schema_dir=SCHEMA_DIR, **kwargs,
):
    """
    Parallel version of make_tables_from_sql

    :type tables: Iterable[str]
    :param tables: An iterable of BigQuery table names
    :type courses: Iterable[str]
    :param courses: An iterable of course IDs
    :type project: str
    :param project: The GCP project against which queries are run
    :type append: bool
    :param append: Whether to append query results to the target tables
    :type query_dir: str
    :param query_dir: The directories where the SQL query files are found
    :type wait: bool
    :param wait: Whether to wait for the BigQuery load jobs to complete
    :type geo_table: str
    :param geo_table: Table name in BigQuery with geolocation data for IPs
    :type youtube_table: str
    :param youtube_table: Table name in BigQuery with YouTube video details
    :type safile: Union[None, str]
    :param safile: GCP service account file to use to connect to BigQuery
    :type size: int
    :param size: Size of the process pool to run queries in parallel
    :type logger: logging.Logger
    :param logger: A Logger object with which to report steps carried out
    :type fail_fast: bool
    :param fail_fast: Whether to stop processing after the first error
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :rtype: Dict[str, Dict[str, Dict[str, str]]]
    :return: A dict mapping course_ids to tables and their query errors
    """
    query_dir = query_dir or QUERY_DIR
    schema_dir = schema_dir or SCHEMA_DIR
    # Don't get extra_args overwrite the given course ID
    kwargs.pop('course_id', None)
    if len(courses) < size:
        size = len(courses)
    results = dict()
    with ProcessPool(
        size, initializer=_report_pool_init,
        initargs=(project, safile)
    ) as pool:
        for course_id in courses:
            if logger:
                logger.info(
                    'Making secondary tables for course ID {cid}'.format(
                        cid=course_id
                    )
                )
            kwds = dict(
                tables=tables, course_id=course_id, client=None,
                project=project, append=append, query_dir=query_dir, wait=wait,
                geo_table=geo_table, youtube_table=youtube_table,
                parallel=True, fail_fast=fail_fast, schema_dir=schema_dir,
            )
            kwds.update(kwargs)
            async_result = pool.apply_async(
                func=make_tables_from_sql, kwds=kwds
            )
            results[course_id] = async_result
        for course_id in results:
            result = results[course_id]
            results[course_id] = result.get()
            if logger:
                logger.info(
                    'All queries submitted for course ID {cid}'.format(
                        cid=course_id
                    )
                )
    return results



"""
Module to process SQL files from edX
"""
import os
import signal
import sys
import traceback
import zipfile
from multiprocessing.pool import (
    Pool as ProcessPool, ThreadPool, TimeoutError
)

from simeon.download.utilities import (
    decrypt_files, format_sql_filename
)
from simeon.exceptions import (
    DecryptionError, EarlyExitError, SplitException
)


SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)


proc_zfile = None


def _pool_initializer(fname):
    """
    Process pool initializer
    """
    global proc_zfile
    proc_zfile = zipfile.ZipFile(fname)
    def sighandler(sig, frame):
        raise EarlyExitError(
            'SQL bundle splitting interrupted prematurely'
        )
    sigs = [signal.SIGABRT, signal.SIGTERM, signal.SIGINT]
    for sig in sigs:
        signal.signal(sig, signal.SIG_DFL)


def _batch_them(items, size):
    """
    Batch the given items by size
    """
    bucket = []
    for item in items:
        bucket.append(item)
        if len(bucket) == size:
            yield bucket[:]
            bucket = []
    if bucket:
        yield bucket


def _batch_archive_names(names, size, include_edge=False):
    """
    Batch the file names inside the archive by size
    """
    bucket = []
    for name in names:
        if not include_edge and '-edge' in name:
            continue
        if name.endswith('/'):
            continue
        bucket.append(name)
        if len(bucket) == size:
            yield bucket[:]
            bucket = []
    if bucket:
        yield bucket


def _delete_all(items):
    """
    Delete the given items from the local file system
    """
    for item in items:
        try:
            os.remove(item)
        except:
            continue


def batch_decrypt_files(
    all_files, size=100, verbose=False, logger=None,
    timeout=None, keepfiles=False,
):
    """
    Batch the files by the given size and pass each batch to gpg to decrypt.

    :type all_files: List[str]
    :param all_files: List of file names
    :type size: int
    :param size: The batch size
    :type verbose: bool
    :param verbose: Print the command to be run
    :type logger: logging.Logger
    :param logger: A logging.Logger object to print the command with
    :type timeout: Union[int, None]
    :param timeout: Number of seconds to wait for the decryption to finish
    :type keepfiles: bool
    :param keepfiles: Keep the encrypted files after decrypting them.
    :rtype: None
    :return: Nothing, but decrypts the .sql files from the given archive
    """
    with ThreadPool(5) as pool:
        results = dict()
        for batch in _batch_them(all_files, size):
            async_result = pool.apply_async(
                    func=decrypt_files, kwds=dict(
                        fnames=batch, verbose=verbose,
                        logger=logger, timeout=timeout
                    )
            )
            results[async_result] = batch
        failures = 0
        for result in results:
            try:
                result.get()
            except DecryptionError as excp:
                failures += 1
                if logger:
                    logger.error(excp)
    if failures:
        raise DecryptionError(
            'Multiple files failed to decrypt. Please consult the logs.'
        )
    if not keepfiles:
        _delete_all(all_files)


def unpacker(fname, names, ddir, courses=None, tables_only=False):
    """
    A worker callable to pass a Thread or Process pool
    """
    global proc_zfile
    targets = []
    for name in names:
        name, target_name = format_sql_filename(name)
        if name is None or target_name is None:
            continue
        target_name = os.path.join(ddir, target_name)
        target_dir = os.path.dirname(target_name)
        if target_dir.count('ora'):
            cfolder = os.path.basename(os.path.dirname(target_dir))
        else:
            cfolder = os.path.basename(target_dir)
        if courses and cfolder not in courses:
            continue
        if 'ccx' in cfolder:
            dir_segments = cfolder.replace('-', '__', 1).split('-')
            clean = '{f}__{s}'.format(
                f='_'.join(dir_segments[:-3]), s='_'.join(dir_segments[-3:])
            )
        else:
            clean = (
                '__'.join(cfolder.replace('-', '__', 1).rsplit('-', 1))
                .replace('-', '_')
            )
        clean = clean.replace('.', '_')
        target_dir = target_dir.replace(cfolder, clean)
        target_name = target_name.replace(cfolder, clean)
        if tables_only:
            targets.append(target_name)
            continue
        os.makedirs(target_dir, exist_ok=True)
        with proc_zfile.open(name) as zh, open(target_name, 'wb') as fh:
            for line in zh:
                fh.write(line)
        targets.append(target_name)
    return targets


def process_sql_archive(
    archive, ddir=None, include_edge=False, courses=None,
    size=5, tables_only=False, debug=False,
):
    """
    Unpack and decrypt files inside the given archive

    :type archive: str
    :param archive: SQL data package (a ZIP archive)
    :type ddir: str
    :param ddir: The destination directory of the unpacked files
    :type include: bool
    :param include_edge: Include the files from the edge site
    :type courses: Union[Iterable[str], None]
    :param courses: A list of course IDs whose data files are unpacked
    :type size: int
    :param size: The size of the thread or process pool doing the unpacking
    :type tables_only: bool
    :param tables_only: Whether to extract file names only (no unarchiving)
    :type debug: bool
    :param debug: Show the stacktrace when an error occurs
    :rtype: Iterable[str]
    :return: List of file names
    """
    cpaths = set()
    for c in (courses or []):
        cpaths.add(c.replace('/', '-'))
    if ddir is None:
        ddir, _ = os.path.split(archive)
    out = []
    with zipfile.ZipFile(archive) as zf:
        names = zf.namelist()
    batches = _batch_archive_names(names, len(names) // size, include_edge)
    with ProcessPool(
        size, initializer=_pool_initializer,
        initargs=(archive,)
    ) as pool:
        results = []
        for batch in batches:
            results.append(pool.apply_async(
                unpacker, args=(archive, batch, ddir, cpaths, tables_only)
            ))
        processed = 0
        while processed < len(results):
            for result in results:
                try:
                    targets = result.get(timeout=60)
                    processed += 1
                    out.extend(targets or [])
                except TimeoutError:
                    continue
                except KeyboardInterrupt:
                    raise SplitException(
                        'The SQL bundle unpacking was interrupted by '
                        'the user.'
                    )
                except:
                    _, excp, tb = sys.exc_info()
                    traces = ['{e}'.format(e=excp)]
                    if debug:
                        traces += map(str.strip, traceback.format_tb(tb))
                        excp = '\n'.join(traces)
                    msg = 'Failed to unpack items from archive {a}: {e}'
                    raise SplitException(msg.format(a=archive, e=excp))
    return out



"""
simeon-geoip is a companion script to the simeon tool that helps with
extracting geolocation information from a MaxMind database, and merging
the generated data file to a target table in BigQuery.

To extract geolocation information for IP addresses in a given file, you
can invoke it as follows:

simeon-geoip extract -t -o geo.json.gz ${maxmind_db} ${tracking_log_files}

The above line should generate geolocation data and put it in geo.json.gz
using the given tracking log files and the version 2 MaxMind DB.

To merge the generated file to a target BigQuery table, invoke it as follows:

simeon-geoip merge -t ${project}.${dataset}.${geotable} -p ${project} \\
    -S ${service_account_file} geo.json.gz

The above command will merge the given file into the target table given by the
-t option. If the target table does not exist, then the given is loaded directly
into it.
"""
import csv
import glob
import gzip
import json
import os
import sys
import urllib.request as requests
from argparse import (
    ArgumentParser, FileType, RawDescriptionHelpFormatter
)
from datetime import datetime

import simeon
import simeon.scripts.utilities as cli_utils
import simeon.upload.gcp as gcp

try:
    from geoip2.database import Reader as GeoReader
except ImportError:
    GeoReader = None


SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)


def import_un_denominations(fname=None):
    """
    Import UN denominations data from GitHub if fname is
    not provided.
    Generate records with said data by mapping country 2-letter ISO codes
    to their UN denominations.
    """
    header = [
        'iso_code', 'code', 'name', 'un_major_region',
        'continent', 'un_economic_group', 'un_developing_nation',
        'un_special_region'
    ]
    if fname:
        fp = open(fname)
        next(fp)
    else:
        url = (
            'https://raw.githubusercontent.com/mitodl/'
            'world_geographic_regions/master/'
            'geographic_regions_by_country.csv'
        )
        req = requests.Request(url, method='GET')
        resp = requests.urlopen(req)
        fp = resp.fp
        next(fp)
        fp = (l.decode('utf8', 'ignore') for l in fp)
    reader = csv.DictReader(fp, delimiter=',', fieldnames=header)
    out = dict()
    for row in reader:
        out[row.get('iso_code')] = dict((k, row[k]) for k in header[3:])
    return out


def make_geo_data(
    db, ip_files, outfile='geoip.json.gz',
    un_data=None, tracking_logs=False, logger=None,
):
    """
    Given a MaxMind DB and a credentials file to PROPROD,
    extract distinct IPs from obscured and get their location data

    :type fname: geoip2.database.Reader
    :param fname: MaxMind geolocation database Reader object to extract info
    :type ip_file: Iterable[str]
    :param ip_file: List of files (text or tracking log) containing IPs
    :type outfile: str
    :param outfile: File in which to dump geolocation data
    :type un_data: Union[Dict[str, str], None]
    :param un_data: Dictionary containing UN country denomination information
    :type tracking_logs: bool
    :param tracking_logs: Whether or not the given IP files are tracking logs
    :type logger: Union[logging.Logger, None]
    :param logger: A logger object to log messages to specific streams
    :rtype: None
    :returns: Writes data to the given output file in append mode
    """
    with gzip.open(outfile, 'at') as outh:
        for ip_file in ip_files:
            if tracking_logs:
                fh = gzip.open(ip_file, 'rt')
                reader = map(json.loads, fh)
            else:
                fh = open(ip_file)
                reader = csv.DictReader(fh, fieldnames=['ip'])
            seen = set()
            line = 0
            while True:
                line += 1
                try:
                    rec = next(reader)
                except StopIteration:
                    break
                except Exception as excp:
                    msg = (
                        'Record {i} from {f} could not be parsed: {e}. '
                        'Skipping it...'
                    ).format(i=line, f=ip_file, e=excp)
                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg, file=sys.stderr)
                    continue
                ip_address = rec.get('ip')
                if not ip_address or ip_address in seen:
                    continue
                seen.add(ip_address)
                try:
                    info = db.city(ip_address)
                    un_info = un_data.get(info.country.iso_code, {})
                    subdiv = info.subdivisions.most_specific
                    row = {
                        'ip': ip_address,
                        'city': info.city.names.get('en'),
                        'countryLabel': info.country.names.get('en'),
                        'country': info.country.iso_code,
                        'cc_by_ip': info.country.iso_code,
                        'postalCode': info.postal.code,
                        'continent': info.continent.names.get('en'),
                        'subdivision': subdiv.names.get('en'),
                        'region': subdiv.iso_code,
                        'latitude': info.location.latitude,
                        'longitude': info.location.longitude,
                    }
                    row.update(un_info)
                except Exception:
                    row = {
                        'ip': ip_address,
                    }
                outh.write(json.dumps(row) + '\n')
            fh.close()


def main():
    """
    geoip entry point
    """
    parser = ArgumentParser(
        description=__doc__,
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '--log-file', '-L',
        help='Log file to use when simeon prints messages. Default: stdout',
        type=FileType('a'),
        default=sys.stdout,
    )
    parser.add_argument(
        '--quiet', '-Q',
        help='Only print error messages to standard streams.',
        action='store_false',
        dest='verbose',
    )
    parser.add_argument(
        '--config-file', '-C',
        help=(
            'The INI configuration file to use for default arguments.'
        ),
    )
    parser.add_argument(
        '--version', '-v',
        action='version',
        version='%(prog)s {v}'.format(v=simeon.__version__)
    )
    subparsers = parser.add_subparsers(
        description='Choose a subcommand to carry out a task with simeon-geoip',
        dest='command'
    )
    subparsers.required = True
    extracter = subparsers.add_parser(
        'extract',
        help='Extract geolocation information from the given MaxMind DB',
        description=(
            'Extract geolocation information for the IP addresses '
            'in the given IP files from the MaxMind version2 database'
        )
    )
    extracter.add_argument(
        'db',
        help='MaxMind version 2 geolocation database file.'
    )
    extracter.add_argument(
        'ip_files',
        help=(
            'File(s) containing IP addresses. If it\'s a text file, '
            'it should be one IP address per record.\n\nIf the file is '
            'a tracking log file, then each record is assumed to have '
            'an "ip" string value.'
        ),
        nargs='+',
    )
    extracter.add_argument(
        '--un-data', '-u',
        help=(
            'Path to a file with UN denominations. If no file is provided, '
            'one is downloaded from Github at '
            'https://raw.githubusercontent.com/mitodl/'
            'world_geographic_regions/master/'
            'geographic_regions_by_country.csv'
        ),
    )
    extracter.add_argument(
        '--output', '-o',
        help='Output file name for the generated data. Default: %(default)s',
        default=os.path.join(
            os.getcwd(),
            'geoip_{dt}.json.gz'.format(
                dt=datetime.now().strftime('%Y%m%d%H%M%S')
            )
        ),
    )
    extracter.add_argument(
        '--tracking-logs', '-t',
        help='Whether or not the given ip files are tracking log files',
        action='store_true',
    )
    merger = subparsers.add_parser(
        'merge',
        help='Merge the given file to a target BigQuery table name',
        description=(
            'Merge the given file to a target BigQuery table name'
        )
    )
    merger.add_argument(
        'geofile',
        help='A .json.gz file generated from the extract command'
    )
    merger.add_argument(
        '--project', '-p',
        help='The BigQuery project id where the target table resides.'
    )
    merger.add_argument(
        '--service-account-file', '-S',
        help='The service account file to use when connecting to BigQuery'
    )
    merger.add_argument(
        '--geo-table', '-g',
        help='The target table where the geolocation data are stored.',
        default='geocode.geoip',
        type=cli_utils.bq_table,
    )
    merger.add_argument(
        '--column', '-c',
        help=(
            'The column on which to to merge the file and table. '
            'Default: %(default)s'
        ),
        default='ip',
    )
    merger.add_argument(
        '--schema-dir', '-s',
        help=(
            'Directory where schema file are found. '
            'Default: {d}'.format(d=SCHEMA_DIR)
        ),
    )
    merger.add_argument(
        '--update-description', '-u',
        help=(
            'Update the description of the destination table with '
            'the "description" value from the corresponding schema file'
        ),
        action='store_true',
    )
    args = parser.parse_args()
    args.logger = cli_utils.make_logger(
        user='SIMEON-GEOIP:{c}'.format(c=args.command.upper()),
        verbose=args.verbose,
        stream=args.log_file,
    )
    if not GeoReader:
        args.logger.error(
            'simeon was installed without geoip2. '
            'please reinstall it with python -m pip install simeon[geoip]. '
            'Or, install geoip2 with python -m pip install geoip2.'
        )
        sys.exit(1)
    if args.command == 'extract':
        files = []
        for file_ in args.ip_files:
            if '*' in file_:
                files.extend(glob.iglob(file_))
            else:
                files.append(file_)
        args.ip_files = files
        if not args.ip_files:
            msg = 'No valid IP data provided. Exiting...'
            if args.logger:
                args.logger.error(msg)
            else:
                print(msg, file=sys.stderr)
        try:
            os.remove(args.output)
        except OSError:
            pass
        try:
            un_denoms = import_un_denominations(args.un_data)
        except Exception as e:
            msg = 'Failed to get UN denominations because: {e}'
            if args.logger:
                args.logger.warning(msg.format(e=e))
            else:
                print(msg.format(e=e), file=sys.stderr)
            un_denoms = {}
        locs = GeoReader(args.db)
        make_geo_data(
            db=locs, ip_files=args.ip_files,
            outfile=args.output, un_data=un_denoms,
            tracking_logs=args.tracking_logs,
            logger=args.logger,
        )
        args.logger.info('Done processing the given IP files')
    else:
        try:
            configs = cli_utils.find_config(args.config_file)
        except Exception as excp:
            args.logger.error(str(excp).replace('\n', ' '))
            sys.exit(1)
        for k, v in cli_utils.CONFIGS.items():
            for (attr, cgetter) in v:
                cli_arg = getattr(args, attr, None)
                config_arg = cgetter(configs, k, attr, fallback=None)
                if not cli_arg and config_arg:
                    setattr(args, attr, config_arg)
        keys = ('geo-table', 'column', 'project')
        if not all(getattr(args, k.replace('-', '_'), None) for k in keys):
            msg = 'The following options expected valid values: {o}'
            args.logger.error(msg.format(o=', '.join(keys)))
            sys.exit(1)
        args.logger.info(
            'Merging {f} to {t}'.format(f=args.geofile, t=args.geo_table)
        )
        args.logger.info('Connecting to BigQuery')
        try:
            if args.service_account_file is not None:
                client = gcp.BigqueryClient.from_service_account_json(
                    args.service_account_file,
                    project=args.project
                )
            else:
                client = client = gcp.BigqueryClient(
                    project=args.project
                )
        except Exception as excp:
            errmsg = 'Failed to connect to BigQuery: {e}'
            args.logger.error(errmsg.format(e=excp))
            args.logger.error(
                'The error may be from an invalid service account file'
            )
            sys.exit(1)
        args.logger.info('Connection established')
        try:
            client.merge_to_table(
                fname=args.geofile, table=args.geo_table, col=args.column,
                use_storage=args.geofile.startswith('gs://'),
                schema_dir=args.schema_dir, patch=args.update_description,
            )
        except Exception as excp:
            msg = 'Merging {f} to {t} failed with the following: {e}'
            args.logger.error(
                msg.format(f=args.geofile, t=args.geo_table, e=excp)
            )
            sys.exit(1)
        msg = 'Successfully merged the records in {f} to the table {t}'
        args.logger.info(msg.format(f=args.geofile, t=args.geo_table))


if __name__ == '__main__':
    main()



"""
Module to process email opt-in data from edX
"""
import csv
import gzip
import json
import os
import zipfile
from datetime import datetime

from simeon.download.utilities import decrypt_files


SCHEMA_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'upload', 'schemas'
)


def parse_date(datestr):
    """
    Convert datestr to an iso formatted date. If not possible, return None
    """
    try:
        # It's been UTC forever, so there is no need to capture the timezone
        # component of the string
        datestr = datestr.split('+')[0]
        dt = datetime.strptime(datestr, '%Y-%m-%d %H:%M:%S')
        return dt.isoformat()
    except Exception:
        return None


def process_email_file(
    fname, verbose=True, logger=None, timeout=None, keepfiles=False
):
    """
    Email opt-in files are kind of different in that
    they are zip archives inside of which reside GPG encrypted files.

    :type fname: str
    :param fname: Zip archive containing the email opt-in data file
    :type verbose: bool
    :param verbose: Whether to print stuff when decrypting
    :type logger: logging.Logger
    :param logger: A Logger object to print messages with
    :type timeout: Union[int, None]
    :param timeout: Number of seconds to wait for the decryption to finish
    :type keepfiles: bool
    :param keepfiles: Whether to keep the .gpg files after decrypting them
    :rtype: str
    :return: Returns the path to the decrypted file
    """
    dirname, out = os.path.split(fname)
    out, _ = os.path.splitext(out)
    out = os.path.join(dirname, '{o}.csv.gpg'.format(o=out))
    with zipfile.ZipFile(fname) as zf, open(out, 'wb') as fh:
        for file_ in zf.infolist():
            if file_.filename.endswith('/'):
                continue
            with zf.open(file_) as zfh:
                while True:
                    chunk = zfh.read(10485760)
                    if not chunk:
                        break
                    fh.write(chunk)
            decrypt_files(
                fnames=out, verbose=verbose,
                logger=logger, timeout=timeout
            )
    if not keepfiles:
        try:
            os.remove(out)
        except OSError:
            pass
    return os.path.splitext(out)[0]


def compress_email_files(files, ddir, schema_dir=SCHEMA_DIR):
    """
    Generate a GZIP JSON file in the given ddir directory
    using the contents of the files.
    :NOTE: schema_dir is not used yet. But we may use to check that
    the generated records match their destination tables.

    :type files: Iterable[str]
    :param files: An iterable of email opt-in CSV files to process
    :type ddir: str
    :param ddir: A destination directory
    :type schema_dir: Union[None, str]
    :param schema_dir: Directory where schema files live
    :rtype: None
    :return: Writes the contents of files into email_opt_in.json.gz
    """
    schema_dir = schema_dir or SCHEMA_DIR
    outname = os.path.join(ddir, 'email_opt_in.json.gz')
    os.makedirs(ddir, exist_ok=True)
    with gzip.open(outname, 'wt') as fh:
        for file_ in files:
            with open(file_) as infile:
                cols = [c.strip() for c in next(infile).split(',')]
                reader = csv.DictReader(
                    infile, delimiter=',',
                    lineterminator='\n', fieldnames=cols
                )
                for row in reader:
                    row['user_id'] = int(row['user_id'])
                    cid = (row.get('course_id') or '').split(':')[-1]
                    row['course_id'] = cid.replace(
                        '+', '/', 2
                    ).replace('+', '_')
                    row['preference_set_datetime'] = parse_date(
                        row.get('preference_set_datetime') or ''
                    )
                    is_opt = (row.get('is_opted_in_for_email') or '').strip()
                    row['is_opted_in_for_email'] = is_opt.lower() == 'true'
                    fh.write(json.dumps(row) + '\n')



"""
Package with module to help generate reports like user_info_combo, person_course, etc.
"""
from .utilities import (
    QUERY_DIR, SCHEMA_DIR,
    make_course_axis, make_forum_table, make_grades_persistent,
    make_grading_policy, make_sql_tables_par, make_sql_tables_seq,
    make_user_info_combo, make_table_from_sql, make_tables_from_sql,
    make_tables_from_sql_par, wait_for_bq_job_ids, wait_for_bq_jobs,
)



"""
Exception classes for the simeon package
"""
class AWSException(Exception):
    """
    Raised when an S3 resource can't be made
    """
    pass


class DecryptionError(Exception):
    """
    Raised when the GPG decryption process fails.
    """
    pass


class MissingSchemaException(Exception):
    """
    Raised when a schema could not be found for a given BigQuery table name
    """
    pass


class SchemaMismatchException(Exception):
    """
    Raised when a record does not match its corresponding schema
    """
    pass


class BigQueryNameException(Exception):
    """
    Raised when a fully qualified table or dataset name
    can't be created.
    """
    pass


class BadSQLFileException(Exception):
    """
    Raised when a SQL file is not in its expected format
    """
    pass


class MissingFileException(Exception):
    """
    Raised when a necessary file is missing
    """


class MissingQueryFileException(Exception):
    """
    Raised when a report table does not have a query file
    in the given query directory.
    """
    pass


class EarlyExitError(Exception):
    """
    Raised when an early exit is requested by the end user of
    the CLI tool
    """


class LoadJobException(Exception):
    """
    Raised from a BigQuery data load job
    """
    pass


class SplitException(Exception):
    """
    Raised when an issue happens during a split operation
    """
    pass


class SQLQueryException(Exception):
    """
    Raised when calling client.query raises an error
    """
